# The contents of this file are automatically written by
# tools/generate_schema_wrapper.py. Do not modify directly.
from . import core
from altair.utils import use_signature
from altair.utils.schemapi import Undefined


class MarkMethodMixin(object):
    """A mixin class that defines mark methods"""

    def mark_area(self, align=Undefined, angle=Undefined, baseline=Undefined, clip=Undefined,
                  color=Undefined, cursor=Undefined, dx=Undefined, dy=Undefined, fill=Undefined,
                  fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                  fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                  limit=Undefined, opacity=Undefined, orient=Undefined, radius=Undefined,
                  shape=Undefined, size=Undefined, stroke=Undefined, strokeDash=Undefined,
                  strokeDashOffset=Undefined, strokeOpacity=Undefined, strokeWidth=Undefined,
                  style=Undefined, tension=Undefined, text=Undefined, theta=Undefined, **kwds):
        """Set the chart's mark to 'area'
    
        For information on additional arguments, see ``alt.MarkDef``
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, clip=clip, color=color, cursor=cursor,
                    dx=dx, dy=dy, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, opacity=opacity, orient=orient, radius=radius,
                    shape=shape, size=size, stroke=stroke, strokeDash=strokeDash,
                    strokeDashOffset=strokeDashOffset, strokeOpacity=strokeOpacity,
                    strokeWidth=strokeWidth, style=style, tension=tension, text=text, theta=theta,
                    **kwds)
        copy = self.copy(deep=True, ignore=['data'])
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="area", **kwds)
        else:
            copy.mark = "area"
        return copy

    def mark_bar(self, align=Undefined, angle=Undefined, baseline=Undefined, clip=Undefined,
                 color=Undefined, cursor=Undefined, dx=Undefined, dy=Undefined, fill=Undefined,
                 fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                 fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                 limit=Undefined, opacity=Undefined, orient=Undefined, radius=Undefined,
                 shape=Undefined, size=Undefined, stroke=Undefined, strokeDash=Undefined,
                 strokeDashOffset=Undefined, strokeOpacity=Undefined, strokeWidth=Undefined,
                 style=Undefined, tension=Undefined, text=Undefined, theta=Undefined, **kwds):
        """Set the chart's mark to 'bar'
    
        For information on additional arguments, see ``alt.MarkDef``
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, clip=clip, color=color, cursor=cursor,
                    dx=dx, dy=dy, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, opacity=opacity, orient=orient, radius=radius,
                    shape=shape, size=size, stroke=stroke, strokeDash=strokeDash,
                    strokeDashOffset=strokeDashOffset, strokeOpacity=strokeOpacity,
                    strokeWidth=strokeWidth, style=style, tension=tension, text=text, theta=theta,
                    **kwds)
        copy = self.copy(deep=True, ignore=['data'])
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="bar", **kwds)
        else:
            copy.mark = "bar"
        return copy

    def mark_line(self, align=Undefined, angle=Undefined, baseline=Undefined, clip=Undefined,
                  color=Undefined, cursor=Undefined, dx=Undefined, dy=Undefined, fill=Undefined,
                  fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                  fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                  limit=Undefined, opacity=Undefined, orient=Undefined, radius=Undefined,
                  shape=Undefined, size=Undefined, stroke=Undefined, strokeDash=Undefined,
                  strokeDashOffset=Undefined, strokeOpacity=Undefined, strokeWidth=Undefined,
                  style=Undefined, tension=Undefined, text=Undefined, theta=Undefined, **kwds):
        """Set the chart's mark to 'line'
    
        For information on additional arguments, see ``alt.MarkDef``
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, clip=clip, color=color, cursor=cursor,
                    dx=dx, dy=dy, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, opacity=opacity, orient=orient, radius=radius,
                    shape=shape, size=size, stroke=stroke, strokeDash=strokeDash,
                    strokeDashOffset=strokeDashOffset, strokeOpacity=strokeOpacity,
                    strokeWidth=strokeWidth, style=style, tension=tension, text=text, theta=theta,
                    **kwds)
        copy = self.copy(deep=True, ignore=['data'])
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="line", **kwds)
        else:
            copy.mark = "line"
        return copy

    def mark_point(self, align=Undefined, angle=Undefined, baseline=Undefined, clip=Undefined,
                   color=Undefined, cursor=Undefined, dx=Undefined, dy=Undefined, fill=Undefined,
                   fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                   fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                   limit=Undefined, opacity=Undefined, orient=Undefined, radius=Undefined,
                   shape=Undefined, size=Undefined, stroke=Undefined, strokeDash=Undefined,
                   strokeDashOffset=Undefined, strokeOpacity=Undefined, strokeWidth=Undefined,
                   style=Undefined, tension=Undefined, text=Undefined, theta=Undefined, **kwds):
        """Set the chart's mark to 'point'
    
        For information on additional arguments, see ``alt.MarkDef``
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, clip=clip, color=color, cursor=cursor,
                    dx=dx, dy=dy, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, opacity=opacity, orient=orient, radius=radius,
                    shape=shape, size=size, stroke=stroke, strokeDash=strokeDash,
                    strokeDashOffset=strokeDashOffset, strokeOpacity=strokeOpacity,
                    strokeWidth=strokeWidth, style=style, tension=tension, text=text, theta=theta,
                    **kwds)
        copy = self.copy(deep=True, ignore=['data'])
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="point", **kwds)
        else:
            copy.mark = "point"
        return copy

    def mark_text(self, align=Undefined, angle=Undefined, baseline=Undefined, clip=Undefined,
                  color=Undefined, cursor=Undefined, dx=Undefined, dy=Undefined, fill=Undefined,
                  fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                  fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                  limit=Undefined, opacity=Undefined, orient=Undefined, radius=Undefined,
                  shape=Undefined, size=Undefined, stroke=Undefined, strokeDash=Undefined,
                  strokeDashOffset=Undefined, strokeOpacity=Undefined, strokeWidth=Undefined,
                  style=Undefined, tension=Undefined, text=Undefined, theta=Undefined, **kwds):
        """Set the chart's mark to 'text'
    
        For information on additional arguments, see ``alt.MarkDef``
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, clip=clip, color=color, cursor=cursor,
                    dx=dx, dy=dy, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, opacity=opacity, orient=orient, radius=radius,
                    shape=shape, size=size, stroke=stroke, strokeDash=strokeDash,
                    strokeDashOffset=strokeDashOffset, strokeOpacity=strokeOpacity,
                    strokeWidth=strokeWidth, style=style, tension=tension, text=text, theta=theta,
                    **kwds)
        copy = self.copy(deep=True, ignore=['data'])
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="text", **kwds)
        else:
            copy.mark = "text"
        return copy

    def mark_tick(self, align=Undefined, angle=Undefined, baseline=Undefined, clip=Undefined,
                  color=Undefined, cursor=Undefined, dx=Undefined, dy=Undefined, fill=Undefined,
                  fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                  fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                  limit=Undefined, opacity=Undefined, orient=Undefined, radius=Undefined,
                  shape=Undefined, size=Undefined, stroke=Undefined, strokeDash=Undefined,
                  strokeDashOffset=Undefined, strokeOpacity=Undefined, strokeWidth=Undefined,
                  style=Undefined, tension=Undefined, text=Undefined, theta=Undefined, **kwds):
        """Set the chart's mark to 'tick'
    
        For information on additional arguments, see ``alt.MarkDef``
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, clip=clip, color=color, cursor=cursor,
                    dx=dx, dy=dy, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, opacity=opacity, orient=orient, radius=radius,
                    shape=shape, size=size, stroke=stroke, strokeDash=strokeDash,
                    strokeDashOffset=strokeDashOffset, strokeOpacity=strokeOpacity,
                    strokeWidth=strokeWidth, style=style, tension=tension, text=text, theta=theta,
                    **kwds)
        copy = self.copy(deep=True, ignore=['data'])
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="tick", **kwds)
        else:
            copy.mark = "tick"
        return copy

    def mark_rect(self, align=Undefined, angle=Undefined, baseline=Undefined, clip=Undefined,
                  color=Undefined, cursor=Undefined, dx=Undefined, dy=Undefined, fill=Undefined,
                  fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                  fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                  limit=Undefined, opacity=Undefined, orient=Undefined, radius=Undefined,
                  shape=Undefined, size=Undefined, stroke=Undefined, strokeDash=Undefined,
                  strokeDashOffset=Undefined, strokeOpacity=Undefined, strokeWidth=Undefined,
                  style=Undefined, tension=Undefined, text=Undefined, theta=Undefined, **kwds):
        """Set the chart's mark to 'rect'
    
        For information on additional arguments, see ``alt.MarkDef``
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, clip=clip, color=color, cursor=cursor,
                    dx=dx, dy=dy, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, opacity=opacity, orient=orient, radius=radius,
                    shape=shape, size=size, stroke=stroke, strokeDash=strokeDash,
                    strokeDashOffset=strokeDashOffset, strokeOpacity=strokeOpacity,
                    strokeWidth=strokeWidth, style=style, tension=tension, text=text, theta=theta,
                    **kwds)
        copy = self.copy(deep=True, ignore=['data'])
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="rect", **kwds)
        else:
            copy.mark = "rect"
        return copy

    def mark_rule(self, align=Undefined, angle=Undefined, baseline=Undefined, clip=Undefined,
                  color=Undefined, cursor=Undefined, dx=Undefined, dy=Undefined, fill=Undefined,
                  fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                  fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                  limit=Undefined, opacity=Undefined, orient=Undefined, radius=Undefined,
                  shape=Undefined, size=Undefined, stroke=Undefined, strokeDash=Undefined,
                  strokeDashOffset=Undefined, strokeOpacity=Undefined, strokeWidth=Undefined,
                  style=Undefined, tension=Undefined, text=Undefined, theta=Undefined, **kwds):
        """Set the chart's mark to 'rule'
    
        For information on additional arguments, see ``alt.MarkDef``
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, clip=clip, color=color, cursor=cursor,
                    dx=dx, dy=dy, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, opacity=opacity, orient=orient, radius=radius,
                    shape=shape, size=size, stroke=stroke, strokeDash=strokeDash,
                    strokeDashOffset=strokeDashOffset, strokeOpacity=strokeOpacity,
                    strokeWidth=strokeWidth, style=style, tension=tension, text=text, theta=theta,
                    **kwds)
        copy = self.copy(deep=True, ignore=['data'])
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="rule", **kwds)
        else:
            copy.mark = "rule"
        return copy

    def mark_circle(self, align=Undefined, angle=Undefined, baseline=Undefined, clip=Undefined,
                    color=Undefined, cursor=Undefined, dx=Undefined, dy=Undefined, fill=Undefined,
                    fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                    fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                    limit=Undefined, opacity=Undefined, orient=Undefined, radius=Undefined,
                    shape=Undefined, size=Undefined, stroke=Undefined, strokeDash=Undefined,
                    strokeDashOffset=Undefined, strokeOpacity=Undefined, strokeWidth=Undefined,
                    style=Undefined, tension=Undefined, text=Undefined, theta=Undefined, **kwds):
        """Set the chart's mark to 'circle'
    
        For information on additional arguments, see ``alt.MarkDef``
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, clip=clip, color=color, cursor=cursor,
                    dx=dx, dy=dy, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, opacity=opacity, orient=orient, radius=radius,
                    shape=shape, size=size, stroke=stroke, strokeDash=strokeDash,
                    strokeDashOffset=strokeDashOffset, strokeOpacity=strokeOpacity,
                    strokeWidth=strokeWidth, style=style, tension=tension, text=text, theta=theta,
                    **kwds)
        copy = self.copy(deep=True, ignore=['data'])
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="circle", **kwds)
        else:
            copy.mark = "circle"
        return copy

    def mark_square(self, align=Undefined, angle=Undefined, baseline=Undefined, clip=Undefined,
                    color=Undefined, cursor=Undefined, dx=Undefined, dy=Undefined, fill=Undefined,
                    fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                    fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                    limit=Undefined, opacity=Undefined, orient=Undefined, radius=Undefined,
                    shape=Undefined, size=Undefined, stroke=Undefined, strokeDash=Undefined,
                    strokeDashOffset=Undefined, strokeOpacity=Undefined, strokeWidth=Undefined,
                    style=Undefined, tension=Undefined, text=Undefined, theta=Undefined, **kwds):
        """Set the chart's mark to 'square'
    
        For information on additional arguments, see ``alt.MarkDef``
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, clip=clip, color=color, cursor=cursor,
                    dx=dx, dy=dy, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, opacity=opacity, orient=orient, radius=radius,
                    shape=shape, size=size, stroke=stroke, strokeDash=strokeDash,
                    strokeDashOffset=strokeDashOffset, strokeOpacity=strokeOpacity,
                    strokeWidth=strokeWidth, style=style, tension=tension, text=text, theta=theta,
                    **kwds)
        copy = self.copy(deep=True, ignore=['data'])
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="square", **kwds)
        else:
            copy.mark = "square"
        return copy

    def mark_geoshape(self, align=Undefined, angle=Undefined, baseline=Undefined, clip=Undefined,
                      color=Undefined, cursor=Undefined, dx=Undefined, dy=Undefined, fill=Undefined,
                      fillOpacity=Undefined, filled=Undefined, font=Undefined, fontSize=Undefined,
                      fontStyle=Undefined, fontWeight=Undefined, href=Undefined, interpolate=Undefined,
                      limit=Undefined, opacity=Undefined, orient=Undefined, radius=Undefined,
                      shape=Undefined, size=Undefined, stroke=Undefined, strokeDash=Undefined,
                      strokeDashOffset=Undefined, strokeOpacity=Undefined, strokeWidth=Undefined,
                      style=Undefined, tension=Undefined, text=Undefined, theta=Undefined, **kwds):
        """Set the chart's mark to 'geoshape'
    
        For information on additional arguments, see ``alt.MarkDef``
        """
        kwds = dict(align=align, angle=angle, baseline=baseline, clip=clip, color=color, cursor=cursor,
                    dx=dx, dy=dy, fill=fill, fillOpacity=fillOpacity, filled=filled, font=font,
                    fontSize=fontSize, fontStyle=fontStyle, fontWeight=fontWeight, href=href,
                    interpolate=interpolate, limit=limit, opacity=opacity, orient=orient, radius=radius,
                    shape=shape, size=size, stroke=stroke, strokeDash=strokeDash,
                    strokeDashOffset=strokeDashOffset, strokeOpacity=strokeOpacity,
                    strokeWidth=strokeWidth, style=style, tension=tension, text=text, theta=theta,
                    **kwds)
        copy = self.copy(deep=True, ignore=['data'])
        if any(val is not Undefined for val in kwds.values()):
            copy.mark = core.MarkDef(type="geoshape", **kwds)
        else:
            copy.mark = "geoshape"
        return copy


class ConfigMethodMixin(object):
    """A mixin class that defines config methods"""

    @use_signature(core.Config)
    def configure(self, *args, **kwargs):
        copy = self.copy()
        copy.config = core.Config(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_area(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["area"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.AxisConfig)
    def configure_axis(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["axis"] = core.AxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgAxisConfig)
    def configure_axisBand(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["axisBand"] = core.VgAxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgAxisConfig)
    def configure_axisBottom(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["axisBottom"] = core.VgAxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgAxisConfig)
    def configure_axisLeft(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["axisLeft"] = core.VgAxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgAxisConfig)
    def configure_axisRight(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["axisRight"] = core.VgAxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgAxisConfig)
    def configure_axisTop(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["axisTop"] = core.VgAxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgAxisConfig)
    def configure_axisX(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["axisX"] = core.VgAxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgAxisConfig)
    def configure_axisY(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["axisY"] = core.VgAxisConfig(*args, **kwargs)
        return copy

    @use_signature(core.BarConfig)
    def configure_bar(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["bar"] = core.BarConfig(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_circle(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["circle"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.Datasets)
    def configure_datasets(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["datasets"] = core.Datasets(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_geoshape(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["geoshape"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.LegendConfig)
    def configure_legend(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["legend"] = core.LegendConfig(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_line(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["line"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_mark(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["mark"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.Padding)
    def configure_padding(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["padding"] = core.Padding(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_point(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["point"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.ProjectionConfig)
    def configure_projection(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["projection"] = core.ProjectionConfig(*args, **kwargs)
        return copy

    @use_signature(core.RangeConfig)
    def configure_range(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["range"] = core.RangeConfig(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_rect(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["rect"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_rule(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["rule"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.ScaleConfig)
    def configure_scale(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["scale"] = core.ScaleConfig(*args, **kwargs)
        return copy

    @use_signature(core.SelectionConfig)
    def configure_selection(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["selection"] = core.SelectionConfig(*args, **kwargs)
        return copy

    @use_signature(core.MarkConfig)
    def configure_square(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["square"] = core.MarkConfig(*args, **kwargs)
        return copy

    @use_signature(core.StackOffset)
    def configure_stack(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["stack"] = core.StackOffset(*args, **kwargs)
        return copy

    @use_signature(core.StyleConfigIndex)
    def configure_style(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["style"] = core.StyleConfigIndex(*args, **kwargs)
        return copy

    @use_signature(core.TextConfig)
    def configure_text(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["text"] = core.TextConfig(*args, **kwargs)
        return copy

    @use_signature(core.TickConfig)
    def configure_tick(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["tick"] = core.TickConfig(*args, **kwargs)
        return copy

    @use_signature(core.VgTitleConfig)
    def configure_title(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["title"] = core.VgTitleConfig(*args, **kwargs)
        return copy

    @use_signature(core.ViewConfig)
    def configure_view(self, *args, **kwargs):
        copy = self.copy(deep=False)
        if copy.config is Undefined:
            copy.config = core.Config()
        else:
            copy.config = copy.config.copy(deep=False)
        copy.config["view"] = core.ViewConfig(*args, **kwargs)
        return copy


"""
Simple Bar Chart
================
This example shows a basic bar chart created with Altair.
"""
# category: basic charts
from altair import Chart
import pandas as pd

data = pd.DataFrame({
    'a': ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I'],
    'b': [28, 55, 43, 91, 81, 53, 19, 87, 52]
})

chart = Chart(data).mark_bar().encode(
    x='a',
    y='b'
)



"""
A Simple server used to show altair graphics from a prompt or script.

This is adapted from the mpld3 package; see
https://github.com/mpld3/mpld3/blob/master/mpld3/_server.py
"""
import sys
import threading
import webbrowser
import socket
import itertools
import random
from ._py3k_compat import server, IO

JUPYTER_WARNING = """
Note: if you're in the Jupyter notebook, Chart.serve() is not the best
      way to view plots. Consider using Chart.display().
You must interrupt the kernel to cancel this command.
"""



# Mock server used for testing

class MockRequest(object):
    def makefile(self, *args, **kwargs):
        return IO(b"GET /")
    def sendall(self, response):
        pass

class MockServer(object):
    def __init__(self, ip_port, Handler):
        handler = Handler(MockRequest(), ip_port[0], self)

    def serve_forever(self):
        pass

    def server_close(self):
        pass


def generate_handler(html, files=None):
    if files is None:
        files = {}

    class MyHandler(server.BaseHTTPRequestHandler):
        def do_GET(self):
            """Respond to a GET request."""
            if self.path == '/':
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write(html.encode())
            elif self.path in files:
                content_type, content = files[self.path]
                self.send_response(200)
                self.send_header("Content-type", content_type)
                self.end_headers()
                self.wfile.write(content.encode())
            else:
                self.send_error(404)

    return MyHandler


def find_open_port(ip, port, n=50):
    """Find an open port near the specified port"""
    ports = itertools.chain((port + i for i in range(n)),
                            (port + random.randint(-2 * n, 2 * n)))

    for port in ports:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = s.connect_ex((ip, port))
        s.close()
        if result != 0:
            return port
    raise ValueError("no open ports found")


def serve(html, ip='127.0.0.1', port=8888, n_retries=50, files=None,
          jupyter_warning=True, open_browser=True, http_server=None):
    """Start a server serving the given HTML, and (optionally) open a browser

    Parameters
    ----------
    html : string
        HTML to serve
    ip : string (default = '127.0.0.1')
        ip address at which the HTML will be served.
    port : int (default = 8888)
        the port at which to serve the HTML
    n_retries : int (default = 50)
        the number of nearby ports to search if the specified port is in use.
    files : dictionary (optional)
        dictionary of extra content to serve
    jupyter_warning : bool (optional)
        if True (default), then print a warning if this is used within Jupyter
    open_browser : bool (optional)
        if True (default), then open a web browser to the given HTML
    http_server : class (optional)
        optionally specify an HTTPServer class to use for showing the
        figure. The default is Python's basic HTTPServer.
    """
    port = find_open_port(ip, port, n_retries)
    Handler = generate_handler(html, files)

    if http_server is None:
        srvr = server.HTTPServer((ip, port), Handler)
    else:
        srvr = http_server((ip, port), Handler)

    if jupyter_warning:
        try:
            __IPYTHON__
        except:
            pass
        else:
            print(JUPYTER_WARNING)

    # Start the server
    print("Serving to http://{0}:{1}/    [Ctrl-C to exit]".format(ip, port))
    sys.stdout.flush()

    if open_browser:
        # Use a thread to open a web browser pointing to the server
        b = lambda: webbrowser.open('http://{0}:{1}'.format(ip, port))
        threading.Thread(target=b).start()

    try:
        srvr.serve_forever()
    except (KeyboardInterrupt, SystemExit):
        print("\nstopping Server...")

    srvr.server_close()



# The contents of this file are automatically written by
# tools/generate_schema_wrapper.py. Do not modify directly.
import collections
import contextlib
import json

import jsonschema
import six


# If DEBUG_MODE is True, then schema objects are converted to dict and
# validated at creation time. This slows things down, particularly for
# larger specs, but leads to much more useful tracebacks for the user.
# Individual schema classes can override this by setting the
# class-level _class_is_valid_at_instantiation attribute to False
DEBUG_MODE = True


def enable_debug_mode():
    global DEBUG_MODE
    DEBUG_MODE = True


def disable_debug_mode():
    global DEBUG_MODE
    DEBUG_MODE = True


@contextlib.contextmanager
def debug_mode(arg):
    global DEBUG_MODE
    original = DEBUG_MODE
    DEBUG_MODE = arg
    try:
        yield
    finally:
        DEBUG_MODE = original


class SchemaValidationError(jsonschema.ValidationError):
    """A wrapper for jsonschema.ValidationError with friendlier traceback"""
    def __init__(self, obj, err):
        super(SchemaValidationError, self).__init__(**err._contents())
        self.obj = obj

    def __unicode__(self):
        cls = self.obj.__class__
        schema_path = ['{0}.{1}'.format(cls.__module__,cls.__name__)]
        schema_path.extend(self.schema_path)
        schema_path = ' ->'.join(val for val in schema_path[:-1]
                                if val not in ('properties',
                                               'additionalProperties',
                                               'patternProperties'))
        return """Invalid specification

        {0}, validating {1!r}

        {2}
        """.format(schema_path, self.validator, self.message)

    if six.PY3:
        __str__ = __unicode__
    else:
        def __str__(self):
            return unicode(self).encode("utf-8")



class UndefinedType(object):
    """A singleton object for marking undefined attributes"""
    __instance = None
    def __new__(cls, *args, **kwargs):
        if not isinstance(cls.__instance, cls):
            cls.__instance = object.__new__(cls, *args, **kwargs)
        return cls.__instance
    def __repr__(self):
        return 'Undefined'
Undefined = UndefinedType()


class SchemaBase(object):
    """Base class for schema wrappers.

    Each derived class should set the _schema class attribute (and optionally
    the _rootschema class attribute) which is used for validation.
    """
    _schema = None
    _rootschema = None
    _class_is_valid_at_instantiation = True

    def __init__(self, *args, **kwds):
        # Two valid options for initialization, which should be handled by
        # derived classes:
        # - a single arg with no kwds, for, e.g. {'type': 'string'}
        # - zero args with zero or more kwds for {'type': 'object'}
        if self._schema is None:
            raise ValueError("Cannot instantiate object of type {0}: "
                             "_schema class attribute is not defined."
                             "".format(self.__class__))

        if kwds:
            assert len(args) == 0
        else:
            assert len(args) in [0, 1]

        # use object.__setattr__ because we override setattr below.
        object.__setattr__(self, '_args', args)
        object.__setattr__(self, '_kwds', kwds)

        if DEBUG_MODE and self._class_is_valid_at_instantiation:
            _ = self.to_dict(validate=True)

    def copy(self, deep=True, ignore=()):
        """Return a copy of the object

        Parameters
        ----------
        deep : boolean, optional
            if True (default) then return a deep copy of all dict, list, and
            SchemaBase objects within the object structure
        ignore : list, optional
            A list of keys for which the contents should not be copied, but
            only stored by reference.
        """
        def _deep_copy(obj, ignore=()):
            if isinstance(obj, SchemaBase):
                args = tuple(_deep_copy(arg) for arg in obj._args)
                kwds = {k: (_deep_copy(v, ignore=ignore)
                            if k not in ignore else v)
                        for k, v in obj._kwds.items()}
                with debug_mode(False):
                    return obj.__class__(*args, **kwds)
            elif isinstance(obj, list):
                return [_deep_copy(v, ignore=ignore) for v in obj]
            elif isinstance(obj, dict):
                return {k: (_deep_copy(v, ignore=ignore)
                            if k not in ignore else v)
                        for k, v in obj.items()}
            else:
                return obj
        if deep:
            return _deep_copy(self, ignore=ignore)
        else:
            with debug_mode(False):
                return self.__class__(*self._args, **self._kwds)

    def __getattr__(self, attr):
        # reminder: getattr is called after the normal lookups
        if attr in self._kwds:
            return self._kwds[attr]
        else:
            try:
                _getattr = super(SchemaBase, self).__getattr__
            except AttributeError:
                _getattr = super(SchemaBase, self).__getattribute__
            return _getattr(attr)

    def __setattr__(self, item , val):
        self._kwds[item] = val

    def __getitem__(self, item):
        return self._kwds[item]

    def __setitem__(self, item, val):
        self._kwds[item] = val

    def __repr__(self):
        if self._kwds:
            args = ("{0}: {1!r}".format(key, val)
                    for key, val in self._kwds.items()
                    if val is not Undefined)
            args = '\n' + ',\n'.join(args)
            return "{0}({{{1}\n}})".format(self.__class__.__name__,
                                            args.replace('\n', '\n  '))
        else:
            return "{0}({1!r})".format(self.__class__.__name__, self._args[0])

    def to_dict(self, validate=True, ignore=[], context={}):
        """Return a dictionary representation of the object

        Parameters
        ----------
        validate : boolean or string
            If True (default), then validate the output dictionary
            against the schema. If "deep" then recursively validate
            all objects in the spec. This takes much more time, but
            it results in friendlier tracebacks for large objects.
        ignore : list
            A list of keys to ignore. This will *not* passed to child to_dict
            function calls.
        context : dict (optional)
            A context dictionary that will be passed to all child to_dict
            function calls

        Returns
        -------
        dct : dictionary
            The dictionary representation of this object

        Raises
        ------
        jsonschema.ValidationError :
            if validate=True and the dict does not conform to the schema
        """
        sub_validate = 'deep' if validate == 'deep' else False

        def _todict(val):
            if isinstance(val, SchemaBase):
                return val.to_dict(validate=sub_validate, context=context)
            elif isinstance(val, (list, tuple)):
                return [_todict(v) for v in val]
            elif isinstance(val, dict):
                return {k: _todict(v) for k, v in val.items()
                        if v is not Undefined}
            else:
                return val

        if self._args and not self._kwds:
            result = _todict(self._args[0])
        elif not self._args:
            result = _todict({k: v for k, v in self._kwds.items()
                              if k not in ignore})
        else:
            raise ValueError("{0} instance has both a value and properties : "
                             "cannot serialize to dict".format(self.__class__))
        if validate:
            try:
                self.validate(result)
            except jsonschema.ValidationError as err:
                raise SchemaValidationError(self, err)
        return result

    def to_json(self, validate=True, ignore=[], context={},
                indent=2, sort_keys=True, **kwargs):
        """Emit the JSON representation for this object as a string.

        Parameters
        ----------
        validate : boolean or string
            If True (default), then validate the output dictionary
            against the schema. If "deep" then recursively validate
            all objects in the spec. This takes much more time, but
            it results in friendlier tracebacks for large objects.
        ignore : list
            A list of keys to ignore. This will *not* passed to child to_dict
            function calls.
        context : dict (optional)
            A context dictionary that will be passed to all child to_dict
            function calls
        indent : integer, default 2
            the number of spaces of indentation to use
        sort_keys : boolean, default True
            if True, sort keys in the output
        **kwargs
            Additional keyword arguments are passed to ``json.dumps()``

        Returns
        -------
        spec : string
            The JSON specification of the chart object.
        """
        dct = self.to_dict(validate=validate, ignore=ignore, context=context)
        return json.dumps(dct, indent=indent, sort_keys=sort_keys, **kwargs)

    @classmethod
    def _default_wrapper_classes(cls):
        """Return the set of classes used within cls.from_dict()"""
        return SchemaBase.__subclasses__()

    @classmethod
    def from_dict(cls, dct, validate=True, _wrapper_classes=None):
        """Construct class from a dictionary representation

        Parameters
        ----------
        dct : dictionary
            The dict from which to construct the class
        validate : boolean
            If True (default), then validate the input against the schema.
        _wrapper_classes : list (optional)
            The set of SchemaBase classes to use when constructing wrappers
            of the dict inputs. If not specified, the result of
            cls._default_wrapper_classes will be used.

        Returns
        -------
        obj : Schema object
            The wrapped schema

        Raises
        ------
        jsonschema.ValidationError :
            if validate=True and dct does not conform to the schema
        """
        if validate:
            cls.validate(dct)
        if _wrapper_classes is None:
            _wrapper_classes = cls._default_wrapper_classes()
        converter = _FromDict(_wrapper_classes)
        return converter.from_dict(constructor=cls, root=cls,
                                   schema=cls._schema, dct=dct)

    @classmethod
    def from_json(cls, json_string, validate=True, **kwargs):
        """Instantiate the object from a valid JSON string

        Parameters
        ----------
        json_string : string
            The string containing a valid JSON chart specification.
        validate : boolean
            If True (default), then validate the input against the schema.
        **kwargs :
            Additional keyword arguments are passed to json.loads

        Returns
        -------
        chart : Chart object
            The altair Chart object built from the specification.
        """
        dct = json.loads(json_string, **kwargs)
        return cls.from_dict(dct, validate=validate)

    @classmethod
    def validate(cls, instance, schema=None):
        """
        Validate the instance against the class schema in the context of the
        rootschema.
        """
        if schema is None:
            schema = cls._schema
        resolver = jsonschema.RefResolver.from_schema(cls._rootschema or cls._schema)
        return jsonschema.validate(instance, schema, resolver=resolver)

    @classmethod
    def resolve_references(cls, schema):
        """Resolve references of the schema the context of this object's schema"""
        resolver = jsonschema.RefResolver.from_schema(cls._rootschema
                                                      or cls._schema
                                                      or schema)
        while '$ref' in schema:
            ref, schema = resolver.resolve(schema['$ref'])
        return schema

    def __dir__(self):
        return list(self._kwds.keys())


class _FromDict(object):
    """Class used to construct SchemaBase class hierarchies from a dict

    The primary purpose of using this class is to be able to build a hash table
    that maps schemas to their wrapper classes. The candidate classes are
    specified in the ``class_list`` argument to the constructor.
    """
    _hash_exclude_keys = ('definitions', 'title', 'description', '$schema', 'id')

    def __init__(self, class_list):
        # Create a mapping of a schema hash to a list of matching classes
        # This lets us quickly determine the correct class to construct
        self.class_dict = collections.defaultdict(list)
        for cls in class_list:
            if cls._schema is not None:
                self.class_dict[self.hash_schema(cls._schema)].append(cls)

    @classmethod
    def hash_schema(cls, schema, use_json=True):
        """
        Compute a python hash for a nested dictionary which
        properly handles dicts, lists, sets, and tuples.

        At the top level, the function excludes from the hashed schema all keys
        listed in `exclude_keys`.

        This implements two methods: one based on conversion to JSON, and one based
        on recursive conversions of unhashable to hashable types; the former seems
        to be slightly faster in several benchmarks.
        """
        if cls._hash_exclude_keys:
            schema = {key: val for key, val in schema.items()
                      if key not in cls._hash_exclude_keys}
        if use_json:
            s = json.dumps(schema, sort_keys=True)
            return hash(s)
        else:
            def _freeze(val):
                if isinstance(val, dict):
                    return frozenset((k, _freeze(v)) for k, v in val.items())
                elif isinstance(val, set):
                    return frozenset(map(_freeze, val))
                elif isinstance(val, list) or isinstance(val, tuple):
                    return tuple(map(_freeze, val))
                else:
                    return val
            return hash(_freeze(schema))

    @staticmethod
    def _passthrough(*args, **kwds):
        """An object constructor that simply passes arguments through"""
        if kwds and not args:
            return kwds
        elif args and not kwds:
            assert len(args) == 1
            return args[0]
        else:
            raise ValueError("Both args and kwds supplied")

    def from_dict(self, constructor, root, schema, dct):
        """Construct an object from a dict representation"""
        # TODO: introspect lists, objects, etc. when they don't have a wrapper.
        #       could do this by passing the schema rather than cls.
        schema = root.resolve_references(schema)

        def _get_constructor(schema):
            # TODO: do something more than simply selecting the last match?
            hash_ = self.hash_schema(schema)
            matches = self.class_dict[hash_]
            constructor = matches[-1] if matches else self._passthrough
            schema = root.resolve_references(schema)
            return constructor, schema

        if 'anyOf' in schema or 'oneOf' in schema:
            schemas = schema.get('anyOf', []) + schema.get('oneOf', [])
            for this_schema in schemas:
                this_constructor, this_schema = _get_constructor(this_schema)
                try:
                    root.validate(dct, this_schema)
                except jsonschema.ValidationError:
                    continue
                else:
                    return self.from_dict(this_constructor, root, this_schema, dct)

        if isinstance(dct, dict):
            # TODO: handle schemas for additionalProperties/patternProperties
            props = schema.get('properties', {})
            kwds = {}
            for key, val in dct.items():
                if key in props:
                    prop_constructor, prop_schema = _get_constructor(props[key])
                    val = self.from_dict(prop_constructor, root, prop_schema, val)
                kwds[key] = val
            return constructor(**kwds)

        elif isinstance(dct, list):
            if 'items' in schema:
                item_schema = schema['items']
                item_constructor, item_schema = _get_constructor(item_schema)
            else:
                item_schema = {}
                item_constructor = self._passthrough
            dct = [self.from_dict(item_constructor, root, item_schema, val)
                   for val in dct]
            return constructor(dct)
        else:
            return constructor(dct)



"""Unit tests for altair API"""

import io
import json
import os
import tempfile

import pytest
import pandas as pd

import altair.vegalite.v2 as alt


def test_chart_data_types():
    Chart = lambda data: alt.Chart(data).mark_point().encode(x='x:Q', y='y:Q')

    # Url Data
    data = '/path/to/my/data.csv'
    dct = Chart(data).to_dict()
    assert dct['data'] == {'url': data}

    # Dict Data
    data = {"values": [{"x": 1, "y": 2}, {"x": 2, "y": 3}]}
    dct = Chart(data).to_dict()
    assert dct['data'] == data

    # DataFrame data
    data = pd.DataFrame({"x": range(5), "y": range(5)})
    dct = Chart(data).to_dict()
    assert dct['data']['values'] == data.to_dict(orient='records')

    # Altair data object
    data = alt.NamedData(name='Foo')
    dct = Chart(data).to_dict()
    assert dct['data'] == {'name': 'Foo'}


def test_chart_infer_types():
    data = pd.DataFrame({'x': pd.date_range('2012', periods=10, freq='Y'),
                         'y': range(10),
                         'c': list('abcabcabca')})

    def _check_encodings(chart):
        dct = chart.to_dict()
        assert dct['encoding']['x']['type'] == 'temporal'
        assert dct['encoding']['x']['field'] == 'x'
        assert dct['encoding']['y']['type'] == 'quantitative'
        assert dct['encoding']['y']['field'] == 'y'
        assert dct['encoding']['color']['type'] == 'nominal'
        assert dct['encoding']['color']['field'] == 'c'

    # Pass field names by keyword
    chart = alt.Chart(data).mark_point().encode(x='x', y='y', color='c')
    _check_encodings(chart)

    # pass Channel objects by keyword
    chart = alt.Chart(data).mark_point().encode(x=alt.X('x'), y=alt.Y('y'),
                                                color=alt.Color('c'))
    _check_encodings(chart)

    # pass Channel objects by value
    chart = alt.Chart(data).mark_point().encode(alt.X('x'), alt.Y('y'),
                                                alt.Color('c'))
    _check_encodings(chart)

    # override default types
    chart = alt.Chart(data).mark_point().encode(alt.X('x', type='nominal'),
                                                alt.Y('y', type='ordinal'))
    dct = chart.to_dict()
    assert dct['encoding']['x']['type'] == 'nominal'
    assert dct['encoding']['y']['type'] == 'ordinal'


def test_chart_operations():
    data = pd.DataFrame({'x': pd.date_range('2012', periods=10, freq='Y'),
                         'y': range(10),
                         'c': list('abcabcabca')})
    chart1 = alt.Chart(data).mark_line().encode(x='x', y='y', color='c')
    chart2 = chart1.mark_point()
    chart3 = chart1.mark_circle()
    chart4 = chart1.mark_square()

    chart = chart1 + chart2 + chart3
    assert isinstance(chart, alt.LayerChart)
    assert len(chart.layer) == 3
    chart += chart4
    assert len(chart.layer) == 4

    chart = chart1 | chart2 | chart3
    assert isinstance(chart, alt.HConcatChart)
    assert len(chart.hconcat) == 3
    chart |= chart4
    assert len(chart.hconcat) == 4

    chart = chart1 & chart2 & chart3
    assert isinstance(chart, alt.VConcatChart)
    assert len(chart.vconcat) == 3
    chart &= chart4
    assert len(chart.vconcat) == 4


def test_selection_to_dict():
    brush = alt.selection(type='interval')

    # test some value selections
    # Note: X and Y cannot have conditions
    chart_values = alt.Chart('path/to/data.json').mark_point().encode(
        color=alt.condition(brush, alt.ColorValue('red'), alt.ColorValue('blue')),
        opacity=alt.condition(brush, alt.value(0.5), alt.value(1.0)),
        text=alt.condition(brush, alt.TextValue('foo'), alt.value('bar'))
    ).to_dict()

    # test some field selections
    # Note: X and Y cannot have conditions
    # Conditions cannot both be fields
    chart_fields = alt.Chart('path/to/data.json').mark_point().encode(
        color=alt.condition(brush, alt.Color('col1:N'), alt.value('blue')),
        opacity=alt.condition(brush, 'col1:N', alt.value(0.5)),
        text=alt.condition(brush, alt.value('abc'), alt.Text('col2:N')),
        size=alt.condition(brush, alt.value(20), 'col2:N')
    ).to_dict()


@pytest.mark.parametrize('format', ['html', 'json', 'png', 'svg'])
def test_savechart(format):
    from ..examples.bar import chart

    if format in ['html', 'json']:
        out = io.StringIO()
        mode = 'r'
    else:
        out = io.BytesIO()
        mode = 'rb'
    fid, filename = tempfile.mkstemp(suffix='.' + format)

    try:
        # selenium may not be installed; skip the test if we get a selenium error.
        try:
            chart.savechart(out, format=format)
            chart.savechart(filename)
        except RuntimeError as err:
            if format in ['png', 'svg'] and 'selenium' in str(err):
                pytest.skip("selenium installation required for png/svg export")
            else:
                raise

        out.seek(0)
        with open(filename, mode) as f:
            assert f.read() == out.read()
    finally:
        os.remove(filename)

    out.seek(0)

    if format == 'json':
        spec = json.load(out)
        assert '$schema' in spec

    elif format == 'html':
        content = out.read()
        assert content.startswith('\n<!DOCTYPE html>')


def test_facet_parse():
    chart = alt.Chart('data.csv').mark_point().encode(
        x='x:Q',
        y='y:Q'
    ).facet(
        row='row:N',
        column='column:O'
    )
    dct = chart.to_dict()
    assert dct['data'] == {'url': 'data.csv'}
    assert 'data' not in dct['spec']
    assert dct['facet'] == {'column': {'field': 'column', 'type': 'ordinal'},
                            'row': {'field': 'row', 'type': 'nominal'}}



LONG_DESCRIPTION = """
Altair: A declarative statistical visualization library for Python.

http://altair-viz.github.io/

This package provides a Python API for building statistical visualizations
in a declarative manner. This API contains no actual visualization rendering
code, but instead emits JSON data structures following the `Vega-Lite`_
specification. For convenience, Altair can optionally use `ipyvega`_ to
seamlessly display client-side renderings in the Jupyter notebook.

.. image:: https://raw.githubusercontent.com/altair-viz/altair/master/images/cars.png

Please note that if you wish to use altair in the Jupyter Notebook, the
`ipyvega`_ notebook extension must be enabled as follows::

    $ pip install altair
    $ pip install --upgrade notebook
    $ jupyter nbextension install --sys-prefix --py vega

See the `Altair Documentation`_ for tutorials, detailed installation
instructions, and examples.
See the `Altair Github Repository`_ for issues, bug reports, and contributions.

.. _Altair Github Repository: http://github.com/altair-viz/altair/
.. _Altair Documentation: http://altair-viz.github.io/
.. _Vega-Lite: https://github.com/vega/vega-lite
.. _ipyvega: https://github.com/vega/ipyvega
"""

import io
import os
import re

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

#==============================================================================
# Utilities
#==============================================================================

def read(path, encoding='utf-8'):
    path = os.path.join(os.path.dirname(__file__), path)
    with io.open(path, encoding=encoding) as fp:
        return fp.read()


def version(path):
    """Obtain the packge version from a python file e.g. pkg/__init__.py

    See <https://packaging.python.org/en/latest/single_source_version.html>.
    """
    version_file = read(path)
    version_match = re.search(r"""^__version__ = ['"]([^'"]*)['"]""",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

HERE = os.path.abspath(os.path.dirname(__file__))

# From https://github.com/jupyterlab/jupyterlab/blob/master/setupbase.py, BSD licensed
def find_packages(top=HERE):
    """
    Find all of the packages.
    """
    packages = []
    for d, dirs, _ in os.walk(top, followlinks=True):
        if os.path.exists(os.path.join(d, '__init__.py')):
            packages.append(os.path.relpath(d, top).replace(os.path.sep, '.'))
        elif d != top:
            # Do not look for packages in subfolders if current is not a package
            dirs[:] = []
    return packages

#==============================================================================
# Variables
#==============================================================================



DESCRIPTION         = "Altair: A declarative statistical visualization library for Python."
NAME                = "altair"
PACKAGES            = find_packages()
PACKAGE_DATA        = {'altair': [
                                  'vega/v2/schema/*.json',
                                  'vega/v3/schema/*.json',
                                  'vegalite/v1/schema/*.json',
                                  'vegalite/v2/schema/*.json'
                                  ]
                      }
AUTHOR              = "Brian E. Granger & Jake VanderPlas"
AUTHOR_EMAIL        = "jakevdp@gmail.com"
URL                 = 'http://altair-viz.github.io'
DOWNLOAD_URL        = 'http://github.com/altair-viz/altair/'
LICENSE             = 'BSD 3-clause'
INSTALL_REQUIRES    = ['entrypoints',
                       'ipython',
                       'jsonschema',
                       'numpy',
                       'pandas',
                       'pytest',
                       'six',
                       'toolz',
                       'vega_datasets']
VERSION             = version('altair/__init__.py')


setup(name=NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      url=URL,
      download_url=DOWNLOAD_URL,
      license=LICENSE,
      packages=PACKAGES,
      package_data=PACKAGE_DATA,
      install_requires=INSTALL_REQUIRES,
      classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'],
     )



"""
Object for generating Python code calls
"""

class CodeGen(object):
    """Class to generate Python function & Class calls

    Parameters
    ----------
    name : string
        The name of the function or object
    args : list (optional)
        The arguments passed to the function. Each argument should either be
        a string or a CodeGen object.
    kwargs : dict (optional)
        Keyword arguments passed to the function. Each key should be a string;
        Each value should be either a string or CodeGen object.
    methods : list (optional)
        Object methods. Each should be a string or a CodeGen object

    Examples
    --------
    >>> method = CodeGen('bar', ['x', 'y'])
    >>> Foo = CodeGen('Foo', args=[4, 5], methods=[method])
    >>> print(str(Foo))
    Foo(4, 5).bar(x, y)

    >>> code = CodeGen('MyObject', kwargs={'f':Foo, 'flag':True})
    >>> print(str(code))
    MyObject(
        f=Foo(4, 5).bar(x, y),
        flag=True,
    )
    """
    def __init__(self, name, args=None, kwargs=None, methods=None):
        self.name = name
        self.args = (args or [])
        self.kwargs = (kwargs or {})
        self.methods = (methods or [])

    @property
    def num_attributes(self):
        return len(self.args) + len(self.kwargs) + len(self.methods)

    def add_args(self, *args):
        self.args.extend(args)
        return self

    def add_kwargs(self, **kwargs):
        self.kwargs.update(kwargs)
        return self

    def add_methods(self, *methods):
        self.methods.extend(methods)
        return self

    def remove_kwargs(self, *kwds):
        for kwd in kwds:
            self.kwargs.pop(kwd, None)
        return self

    def to_str(self, tablevel=0, tabsize=4):
        """Return a string representation of the code"""
        def get_str(obj, tablevel=tablevel, tabsize=tabsize):
            if isinstance(obj, CodeGen):
                return obj.to_str(tablevel=tablevel, tabsize=tabsize)
            elif isinstance(obj, list):
                return '[{0}]'.format(', '.join(get_str(item,
                                                        tablevel + tabsize)
                                                for item in obj))
            else:
                return str(obj)

        args = [get_str(arg) for arg in self.args]
        kwargs = [((tablevel + tabsize) * ' '
                   + '{0}={1}'.format(k, get_str(v, tablevel + tabsize)))
                  for k, v in sorted(self.kwargs.items())]
        if kwargs:
            kwargs = kwargs + [tablevel * ' ']

        if not kwargs and not args:
            call = '{0}()'.format(self.name)
        elif not kwargs:
            call = '{0}({1})'.format(self.name, ', '.join(args))
        elif not args:
            call = '{0}(\n{1})'.format(self.name, ',\n'.join(kwargs))
        else:
            call = '{0}({1}{2})'.format(self.name, ', '.join(args),
                                        ',\n'.join([''] + kwargs))

        for method in self.methods:
            call += '.{0}'.format(get_str(method))

        return call

    def __str__(self):
        return self.to_str()

    def rename(self, newname):
        """Rename function and return self"""
        self.name = newname
        return self

    def convert_arg_to_method(self, arg, methodname=None, depth=1):
        """Convert an argument (and optionally subarguments) to a method

        Examples
        --------
        >>> encoding = CodeGen('Encoding', kwargs=dict(arg1=1, arg2=2))
        >>> chart = CodeGen('Chart', kwargs=dict(encoding=encoding))
        >>> print(str(chart))
        Chart(
            encoding=Encoding(
                arg1=1,
                arg2=2,
            ),
        )
        >>> chart.convert_arg_to_method('encoding', 'encode')
        >>> print(str(chart))
        Chart().encode(
            arg1=1,
            arg2=2,
        )
        """
        if methodname is None:
            methodname = arg

        def submethods(obj, name, depth):
            """Generate methods, and submethods if depth > 1"""
            if depth > 1:
                # if we're going deeper than one method, then cycle through all
                # arguments and recursively call this function if they are a
                # code object
                for subname, subobj in list(obj.kwargs.items()):
                    if isinstance(subobj, CodeGen):
                        obj.kwargs.pop(subname)
                        subname = '{0}_{1}'.format(name, subname)
                        for method in submethods(subobj, subname, depth - 1):
                            yield method
            if obj.num_attributes > 0:
                # if there are still attributes in the object, yield the
                # top-level method.
                yield obj.rename(name)

        code_obj = self.kwargs.pop(arg, CodeGen(''))
        self.add_methods(*submethods(code_obj, methodname, depth))



"""
Tools for importing dependencies
"""
import importlib


def attempt_import(module, error_message):
    """Attempt import of a dependency

    If the dependency is not available, raise a RuntimeError with the appropriate message

    Parameters
    ----------
    module : string
        the module name
    error_message : string
        the error to raise if the module is not importable

    Returns
    -------
    mod : ModuleType
        the imported module

    Raises
    ------
    RuntimeError
    """
    try:
        return importlib.import_module(module)
    except ImportError:
        raise RuntimeError(error_message)



from typing import Callable, Generic, List, TypeVar, Union, cast

import entrypoints



PluginType = TypeVar('PluginType')


class PluginRegistry(Generic[PluginType]):
    """A registry for plugins.

    This is a plugin registry that allows plugins to be loaded/registered
    in two ways:

    1. Through an explicit call to ``.register(name, value)``.
    2. By looking for other Python packages that are installed and provide
       a setuptools entry point group.

    When you create an instance of this claass, provide the name of the 
    entry point group to use::

        reg = PluginRegister('my_entrypoint_group')

    """

    def __init__(self, entry_point_group: str = '', plugin_type=object) -> None:
        """Create a PluginRegistry for a named entry point group.
        
        Parameters
        ==========
        entry_point_group: str
            The name of the entry point group.
        plugin_type: object
            A type that will optionally be used for runtime type checking of
            loaded plugins using isinstance.
        """
        self.entry_point_group = entry_point_group
        self.plugin_type = plugin_type
        self._active = None     # type: None
        self._plugins = {}      # type: dict
        self._options = {}      # type: dict
    

    def register(self, name: str, value: Union[PluginType,None]) -> PluginType:
        """Register a plugin by name and value.

        This method is used for explicit registration of a plugin and shouldn't be
        used to manage entry point managed plugins, which are auto-loaded.

        Parameters
        ==========
        name: str
            The name of the plugin.
        value: PluginType or None
            The actual plugin object to register or None to unregister that plugin.

        Returns
        =======
        plugin: PluginType
            The plugin that was registered or unregistered.
        """
        if value is None and name in self._plugins:
            return self._plugins.pop(name)
        else:
            assert isinstance(value, self.plugin_type)
            self._plugins[name] = value
            return value


    def names(self) -> List[str]:
        """List the names of the registered and entry points plugins."""
        exts = list(self._plugins.keys())
        more_exts = [ep.name for ep in entrypoints.get_group_all(self.entry_point_group)]
        exts.extend(more_exts)
        return exts


    def enable(self, name: str) -> None:
        """Enable a plugin by name."""
        if name not in self._plugins:
            ep = entrypoints.get_single(self.entry_point_group, name)
            value = cast(PluginType, ep.load())
            assert isinstance(value, self.plugin_type)
            self.register(name, value)
        self._active = self._plugins[name]


    def get(self) -> PluginType:
        """Return the currently active plugin."""
        return self._active




"""
Ploynomial Fit Plot
===================
This example shows how to overlay data with a fitted polynomial
"""

import numpy as np
import pandas as pd
import altair as alt

# Generate some random data
rng = np.random.RandomState(1)
x = rng.rand(40) ** 2
y = 10 - 1. / (x + 0.1) + rng.randn(40)
df = pd.DataFrame({'x':x,'y':y})

# Define the degree of the polynomial fit
degree_list = [1, 3, 5]

# Build a dataframe with the fitted data
poly_data = pd.DataFrame({'xfit':np.linspace(df['x'].min(), df['x'].max(), 500)})

for degree in degree_list:
    poly_data[str(degree)] = np.poly1d(np.polyfit(df['x'], df['y'],degree))(poly_data['xfit'])

# Tidy the dataframe so 'degree' is a variable
poly_data = pd.melt(poly_data,
                    id_vars=['xfit'],
                    value_vars=[str(deg) for deg in degree_list],
                    var_name='degree', value_name='yfit')

# Plot the data points on an interactive axis
points = alt.Chart(df).mark_circle(color='black').encode(
    x=alt.X('x', axis=alt.Axis(title='x')),
    y=alt.Y('y', axis=alt.Axis(title='y')),
).interactive()

# Plot the best fit polynomials
polynomial_fit = alt.Chart(poly_data).mark_line().encode(
    x='xfit',
    y='yfit',
    color='degree'
)

chart = points + polynomial_fit



"""
Bubble Plot
-----------------
This example shows how to make a bubble plot.
"""
# category: basic charts

import altair as alt
from vega_datasets import data

source = data.cars()

chart = alt.Chart(source).mark_point().encode(
    x='Horsepower',
    y='Miles_per_Gallon',
    size='Acceleration'
)



# The contents of this file are automatically written by
# tools/generate_schema_wrapper.py. Do not modify directly.

from altair.utils.schemapi import SchemaBase, Undefined

import os
import json

def load_schema():
    """Load the json schema associated with this module's functions"""
    directory = os.path.dirname(__file__)
    with open(os.path.join(directory, 'vega-schema.json')) as f:
        return json.load(f)


class VegaSchema(SchemaBase):
    @classmethod
    def _default_wrapper_classes(cls):
        return VegaSchema.__subclasses__()


class Root(VegaSchema):
    """Root schema wrapper
    
    allOf(container, Mapping(required=[]))
    """
    _schema = load_schema()
    _rootschema = _schema

    def __init__(self, axes=Undefined, background=Undefined, data=Undefined, height=Undefined,
                 legends=Undefined, marks=Undefined, padding=Undefined, predicates=Undefined,
                 scales=Undefined, scene=Undefined, signals=Undefined, viewport=Undefined,
                 width=Undefined, **kwds):
        super(Root, self).__init__(axes=axes, background=background, data=data, height=height,
                                   legends=legends, marks=marks, padding=padding, predicates=predicates,
                                   scales=scales, scene=scene, signals=signals, viewport=viewport,
                                   width=width, **kwds)


class axis(VegaSchema):
    """axis schema wrapper
    
    Mapping(required=[type, scale])
    
    Attributes
    ----------
    scale : string
    
    type : enum('x', 'y')
    
    format : string
    
    formatType : enum('time', 'utc', 'string', 'number')
    
    grid : boolean
    
    layer : enum('front', 'back')
    
    offset : oneOf(float, Mapping(required=[scale, value]))
    
    orient : enum('top', 'bottom', 'left', 'right')
    
    properties : Mapping(required=[])
    
    subdivide : float
    
    tickPadding : float
    
    tickSize : float
    
    tickSizeEnd : float
    
    tickSizeMajor : float
    
    tickSizeMinor : float
    
    ticks : float
    
    title : string
    
    titleOffset : float
    
    values : List(anyOf(string, float))
    
    """
    _schema = {'$ref': '#/defs/axis'}
    _rootschema = Root._schema

    def __init__(self, scale=Undefined, type=Undefined, format=Undefined, formatType=Undefined,
                 grid=Undefined, layer=Undefined, offset=Undefined, orient=Undefined,
                 properties=Undefined, subdivide=Undefined, tickPadding=Undefined, tickSize=Undefined,
                 tickSizeEnd=Undefined, tickSizeMajor=Undefined, tickSizeMinor=Undefined,
                 ticks=Undefined, title=Undefined, titleOffset=Undefined, values=Undefined, **kwds):
        super(axis, self).__init__(scale=scale, type=type, format=format, formatType=formatType,
                                   grid=grid, layer=layer, offset=offset, orient=orient,
                                   properties=properties, subdivide=subdivide, tickPadding=tickPadding,
                                   tickSize=tickSize, tickSizeEnd=tickSizeEnd,
                                   tickSizeMajor=tickSizeMajor, tickSizeMinor=tickSizeMinor,
                                   ticks=ticks, title=title, titleOffset=titleOffset, values=values,
                                   **kwds)


class background(VegaSchema):
    """background schema wrapper
    
    string
    """
    _schema = {'$ref': '#/defs/background'}
    _rootschema = Root._schema

    def __init__(self, *args):
        super(background, self).__init__(*args)


class data(VegaSchema):
    """data schema wrapper
    
    allOf(Mapping(required=[name]), anyOf(Mapping(required=[name, modify]), 
    oneOf(Mapping(required=[source]), Mapping(required=[values]), Mapping(required=[url]))))
    """
    _schema = {'$ref': '#/defs/data'}
    _rootschema = Root._schema

    def __init__(self, name=Undefined, format=Undefined, modify=Undefined, transform=Undefined, **kwds):
        super(data, self).__init__(name=name, format=format, modify=modify, transform=transform, **kwds)


class legend(VegaSchema):
    """legend schema wrapper
    
    anyOf(Mapping(required=[size]), Mapping(required=[shape]), Mapping(required=[fill]), 
    Mapping(required=[stroke]), Mapping(required=[opacity]))
    
    Attributes
    ----------
    """
    _schema = {'$ref': '#/defs/legend'}
    _rootschema = Root._schema

    def __init__(self, *args, **kwds):
        super(legend, self).__init__(*args, **kwds)


class mark(VegaSchema):
    """mark schema wrapper
    
    Mapping(required=[type])
    
    Attributes
    ----------
    type : enum('rect', 'symbol', 'path', 'arc', 'area', 'line', 'rule', 'image', 'text', 
    'group')
    
    delay : numberValue
    
    ease : enum('linear-in', 'linear-out', 'linear-in-out', 'linear-out-in', 'quad-in', 
    'quad-out', 'quad-in-out', 'quad-out-in', 'cubic-in', 'cubic-out', 'cubic-in-out', 
    'cubic-out-in', 'sin-in', 'sin-out', 'sin-in-out', 'sin-out-in', 'exp-in', 'exp-out', 
    'exp-in-out', 'exp-out-in', 'circle-in', 'circle-out', 'circle-in-out', 'circle-out-in', 
    'bounce-in', 'bounce-out', 'bounce-in-out', 'bounce-out-in')
    
    interactive : boolean
    
    key : string
    
    name : string
    
    properties : anyOf(Mapping(required=[enter]), Mapping(required=[update]))
    
    """
    _schema = {'$ref': '#/defs/mark'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, delay=Undefined, ease=Undefined, interactive=Undefined,
                 key=Undefined, name=Undefined, properties=Undefined, **kwds):
        super(mark, self).__init__(type=type, delay=delay, ease=ease, interactive=interactive, key=key,
                                   name=name, properties=properties, **kwds)


class container(VegaSchema):
    """container schema wrapper
    
    Mapping(required=[])
    
    Attributes
    ----------
    axes : List(axis)
    
    legends : List(legend)
    
    marks : List(oneOf(groupMark, visualMark))
    
    scales : List(scale)
    
    scene : Mapping(required=[])
    
    """
    _schema = {'$ref': '#/defs/container'}
    _rootschema = Root._schema

    def __init__(self, axes=Undefined, legends=Undefined, marks=Undefined, scales=Undefined,
                 scene=Undefined, **kwds):
        super(container, self).__init__(axes=axes, legends=legends, marks=marks, scales=scales,
                                        scene=scene, **kwds)


class groupMark(VegaSchema):
    """groupMark schema wrapper
    
    allOf(Mapping(required=[type]), mark, container)
    """
    _schema = {'$ref': '#/defs/groupMark'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, axes=Undefined, delay=Undefined, ease=Undefined,
                 interactive=Undefined, key=Undefined, legends=Undefined, marks=Undefined,
                 name=Undefined, properties=Undefined, scales=Undefined, scene=Undefined, **kwds):
        super(groupMark, self).__init__(type=type, axes=axes, delay=delay, ease=ease,
                                        interactive=interactive, key=key, legends=legends, marks=marks,
                                        name=name, properties=properties, scales=scales, scene=scene,
                                        **kwds)


class visualMark(VegaSchema):
    """visualMark schema wrapper
    
    allOf(not Mapping(required=[]), mark)
    """
    _schema = {'$ref': '#/defs/visualMark'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, delay=Undefined, ease=Undefined, interactive=Undefined,
                 key=Undefined, name=Undefined, properties=Undefined, **kwds):
        super(visualMark, self).__init__(type=type, delay=delay, ease=ease, interactive=interactive,
                                         key=key, name=name, properties=properties, **kwds)


class modify(VegaSchema):
    """modify schema wrapper
    
    List(oneOf(Mapping(required=[type, signal]), Mapping(required=[type, predicate]), 
    Mapping(required=[type, test])))
    """
    _schema = {'$ref': '#/defs/modify'}
    _rootschema = Root._schema

    def __init__(self, *args):
        super(modify, self).__init__(*args)


class padding(VegaSchema):
    """padding schema wrapper
    
    oneOf(enum('strict', 'auto'), float, Mapping(required=[]))
    """
    _schema = {'$ref': '#/defs/padding'}
    _rootschema = Root._schema

    def __init__(self, *args, **kwds):
        super(padding, self).__init__(*args, **kwds)


class predicate(VegaSchema):
    """predicate schema wrapper
    
    oneOf(Mapping(required=[name, type, operands]), Mapping(required=[name, type, operands]), 
    oneOf(Mapping(required=[range]), Mapping(required=[data, field])))
    """
    _schema = {'$ref': '#/defs/predicate'}
    _rootschema = Root._schema

    def __init__(self, *args, **kwds):
        super(predicate, self).__init__(*args, **kwds)


class rule(VegaSchema):
    """rule schema wrapper
    
    anyOf(Mapping(required=[]), Mapping(required=[]))
    """
    _schema = {'$ref': '#/defs/rule'}
    _rootschema = Root._schema

    def __init__(self, *args, **kwds):
        super(rule, self).__init__(*args, **kwds)


class propset(VegaSchema):
    """propset schema wrapper
    
    Mapping(required=[])
    
    Attributes
    ----------
    align : oneOf(Mapping(required=[rule]), List(allOf(rule, allOf(stringModifiers, 
    oneOf(signal, Mapping(required=[value]), Mapping(required=[field]), 
    Mapping(required=[band]))))), allOf(stringModifiers, oneOf(signal, 
    Mapping(required=[value]), Mapping(required=[field]), Mapping(required=[band]))))
    
    angle : numberValue
    
    baseline : oneOf(Mapping(required=[rule]), List(allOf(rule, allOf(stringModifiers, 
    oneOf(signal, Mapping(required=[value]), Mapping(required=[field]), 
    Mapping(required=[band]))))), allOf(stringModifiers, oneOf(signal, 
    Mapping(required=[value]), Mapping(required=[field]), Mapping(required=[band]))))
    
    clip : booleanValue
    
    cursor : stringValue
    
    dx : numberValue
    
    dy : numberValue
    
    endAngle : numberValue
    
    fill : colorValue
    
    fillOpacity : numberValue
    
    font : stringValue
    
    fontSize : numberValue
    
    fontStyle : stringValue
    
    fontWeight : stringValue
    
    height : numberValue
    
    innerRadius : numberValue
    
    interpolate : oneOf(Mapping(required=[rule]), List(allOf(rule, allOf(stringModifiers, 
    oneOf(signal, Mapping(required=[value]), Mapping(required=[field]), 
    Mapping(required=[band]))))), allOf(stringModifiers, oneOf(signal, 
    Mapping(required=[value]), Mapping(required=[field]), Mapping(required=[band]))))
    
    opacity : numberValue
    
    orient : oneOf(Mapping(required=[rule]), List(allOf(rule, allOf(stringModifiers, 
    oneOf(signal, Mapping(required=[value]), Mapping(required=[field]), 
    Mapping(required=[band]))))), allOf(stringModifiers, oneOf(signal, 
    Mapping(required=[value]), Mapping(required=[field]), Mapping(required=[band]))))
    
    outerRadius : numberValue
    
    path : stringValue
    
    radius : numberValue
    
    shape : anyOf(oneOf(Mapping(required=[rule]), List(allOf(rule, allOf(stringModifiers, 
    oneOf(signal, Mapping(required=[value]), Mapping(required=[field]), 
    Mapping(required=[band]))))), allOf(stringModifiers, oneOf(signal, 
    Mapping(required=[value]), Mapping(required=[field]), Mapping(required=[band])))), 
    stringValue)
    
    size : numberValue
    
    startAngle : numberValue
    
    stroke : colorValue
    
    strokeDash : arrayValue
    
    strokeDashOffset : numberValue
    
    strokeOpacity : numberValue
    
    strokeWidth : numberValue
    
    tension : numberValue
    
    text : stringValue
    
    theta : numberValue
    
    url : stringValue
    
    width : numberValue
    
    x : numberValue
    
    x2 : numberValue
    
    xc : numberValue
    
    y : numberValue
    
    y2 : numberValue
    
    yc : numberValue
    
    """
    _schema = {'$ref': '#/defs/propset'}
    _rootschema = Root._schema

    def __init__(self, align=Undefined, angle=Undefined, baseline=Undefined, clip=Undefined,
                 cursor=Undefined, dx=Undefined, dy=Undefined, endAngle=Undefined, fill=Undefined,
                 fillOpacity=Undefined, font=Undefined, fontSize=Undefined, fontStyle=Undefined,
                 fontWeight=Undefined, height=Undefined, innerRadius=Undefined, interpolate=Undefined,
                 opacity=Undefined, orient=Undefined, outerRadius=Undefined, path=Undefined,
                 radius=Undefined, shape=Undefined, size=Undefined, startAngle=Undefined,
                 stroke=Undefined, strokeDash=Undefined, strokeDashOffset=Undefined,
                 strokeOpacity=Undefined, strokeWidth=Undefined, tension=Undefined, text=Undefined,
                 theta=Undefined, url=Undefined, width=Undefined, x=Undefined, x2=Undefined,
                 xc=Undefined, y=Undefined, y2=Undefined, yc=Undefined, **kwds):
        super(propset, self).__init__(align=align, angle=angle, baseline=baseline, clip=clip,
                                      cursor=cursor, dx=dx, dy=dy, endAngle=endAngle, fill=fill,
                                      fillOpacity=fillOpacity, font=font, fontSize=fontSize,
                                      fontStyle=fontStyle, fontWeight=fontWeight, height=height,
                                      innerRadius=innerRadius, interpolate=interpolate, opacity=opacity,
                                      orient=orient, outerRadius=outerRadius, path=path, radius=radius,
                                      shape=shape, size=size, startAngle=startAngle, stroke=stroke,
                                      strokeDash=strokeDash, strokeDashOffset=strokeDashOffset,
                                      strokeOpacity=strokeOpacity, strokeWidth=strokeWidth,
                                      tension=tension, text=text, theta=theta, url=url, width=width,
                                      x=x, x2=x2, xc=xc, y=y, y2=y2, yc=yc, **kwds)


class signal(VegaSchema):
    """signal schema wrapper
    
    Mapping(required=[name])
    
    Attributes
    ----------
    name : not Mapping(required=[])
    
    expr : string
    
    init : Mapping(required=[])
    
    scale : scopedScale
    
    streams : streams
    
    verbose : boolean
    
    """
    _schema = {'$ref': '#/defs/signal'}
    _rootschema = Root._schema

    def __init__(self, name=Undefined, expr=Undefined, init=Undefined, scale=Undefined,
                 streams=Undefined, verbose=Undefined, **kwds):
        super(signal, self).__init__(name=name, expr=expr, init=init, scale=scale, streams=streams,
                                     verbose=verbose, **kwds)


class spec(VegaSchema):
    """spec schema wrapper
    
    allOf(container, Mapping(required=[]))
    """
    _schema = {'$ref': '#/defs/spec'}
    _rootschema = Root._schema

    def __init__(self, axes=Undefined, background=Undefined, data=Undefined, height=Undefined,
                 legends=Undefined, marks=Undefined, padding=Undefined, predicates=Undefined,
                 scales=Undefined, scene=Undefined, signals=Undefined, viewport=Undefined,
                 width=Undefined, **kwds):
        super(spec, self).__init__(axes=axes, background=background, data=data, height=height,
                                   legends=legends, marks=marks, padding=padding, predicates=predicates,
                                   scales=scales, scene=scene, signals=signals, viewport=viewport,
                                   width=width, **kwds)


class streams(VegaSchema):
    """streams schema wrapper
    
    List(Mapping(required=[type, expr]))
    """
    _schema = {'$ref': '#/defs/streams'}
    _rootschema = Root._schema

    def __init__(self, *args):
        super(streams, self).__init__(*args)


class aggregateTransform(VegaSchema):
    """aggregateTransform schema wrapper
    
    Mapping(required=[type])
    Compute summary aggregate statistics
    
    Attributes
    ----------
    type : enum('aggregate')
    
    groupby : List(oneOf(string, signal))
        A list of fields to split the data into groups.
    summarize : oneOf(Mapping(required=[]), List(Mapping(required=[field, ops])))
    
    """
    _schema = {'$ref': '#/defs/aggregateTransform'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, groupby=Undefined, summarize=Undefined, **kwds):
        super(aggregateTransform, self).__init__(type=type, groupby=groupby, summarize=summarize, **kwds)


class binTransform(VegaSchema):
    """binTransform schema wrapper
    
    Mapping(required=[type, field])
    Bins values into quantitative bins (e.g., for a histogram).
    
    Attributes
    ----------
    field : oneOf(string, signal)
        The name of the field to bin values from.
    type : enum('bin')
    
    base : oneOf(float, signal)
        The number base to use for automatic bin determination.
    div : oneOf(List(float), signal)
        An array of scale factors indicating allowable subdivisions.
    max : oneOf(float, signal)
        The maximum bin value to consider.
    maxbins : oneOf(float, signal)
        The maximum number of allowable bins.
    min : oneOf(float, signal)
        The minimum bin value to consider.
    minstep : oneOf(float, signal)
        A minimum allowable step size (particularly useful for integer values).
    output : Mapping(required=[])
        Rename the output data fields
    step : oneOf(float, signal)
        An exact step size to use between bins. If provided, options such as maxbins will be
         ignored.
    steps : oneOf(List(float), signal)
        An array of allowable step sizes to choose from.
    """
    _schema = {'$ref': '#/defs/binTransform'}
    _rootschema = Root._schema

    def __init__(self, field=Undefined, type=Undefined, base=Undefined, div=Undefined, max=Undefined,
                 maxbins=Undefined, min=Undefined, minstep=Undefined, output=Undefined, step=Undefined,
                 steps=Undefined, **kwds):
        super(binTransform, self).__init__(field=field, type=type, base=base, div=div, max=max,
                                           maxbins=maxbins, min=min, minstep=minstep, output=output,
                                           step=step, steps=steps, **kwds)


class crossTransform(VegaSchema):
    """crossTransform schema wrapper
    
    Mapping(required=[type])
    Compute the cross-product of two data sets.
    
    Attributes
    ----------
    type : enum('cross')
    
    diagonal : oneOf(boolean, signal)
        If false, items along the "diagonal" of the cross-product (those elements with the 
        same index in their respective array) will not be included in the output.
    filter : string
        A string containing an expression (in JavaScript syntax) to filter the resulting 
        data elements.
    output : Mapping(required=[])
        Rename the output data fields
    """
    _schema = {'$ref': '#/defs/crossTransform'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, diagonal=Undefined, filter=Undefined, output=Undefined, **kwds):
        super(crossTransform, self).__init__(type=type, diagonal=diagonal, filter=filter, output=output,
                                             **kwds)


class countpatternTransform(VegaSchema):
    """countpatternTransform schema wrapper
    
    Mapping(required=[type])
    
    Attributes
    ----------
    type : enum('countpattern')
    
    case : oneOf(enum('lower', 'upper', 'none'), signal)
        Text case transformation to apply.
    field : oneOf(string, signal)
        The field containing the text to analyze.
    output : Mapping(required=[])
        Rename the output data fields
    pattern : oneOf(string, signal)
        A regexp pattern for matching words in text.
    stopwords : oneOf(string, signal)
        A regexp pattern for matching stopwords to omit.
    """
    _schema = {'$ref': '#/defs/countpatternTransform'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, case=Undefined, field=Undefined, output=Undefined,
                 pattern=Undefined, stopwords=Undefined, **kwds):
        super(countpatternTransform, self).__init__(type=type, case=case, field=field, output=output,
                                                    pattern=pattern, stopwords=stopwords, **kwds)


class linkpathTransform(VegaSchema):
    """linkpathTransform schema wrapper
    
    Mapping(required=[type])
    Computes a path definition for connecting nodes within a node-link network or tree diagram.
    
    Attributes
    ----------
    type : enum('linkpath')
    
    output : Mapping(required=[])
        Rename the output data fields
    shape : oneOf(enum('line', 'curve', 'cornerX', 'cornerY', 'cornerR', 'diagonalX', 
    'diagonalY', 'diagonalR'), signal)
        The path shape to use
    sourceX : oneOf(string, signal)
        The data field that references the source x-coordinate for this link.
    sourceY : oneOf(string, signal)
        The data field that references the source y-coordinate for this link.
    targetX : oneOf(string, signal)
        The data field that references the target x-coordinate for this link.
    targetY : oneOf(string, signal)
        The data field that references the target y-coordinate for this link.
    tension : oneOf(float, signal)
        A tension parameter for the "tightness" of "curve"-shaped links.
    """
    _schema = {'$ref': '#/defs/linkpathTransform'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, output=Undefined, shape=Undefined, sourceX=Undefined,
                 sourceY=Undefined, targetX=Undefined, targetY=Undefined, tension=Undefined, **kwds):
        super(linkpathTransform, self).__init__(type=type, output=output, shape=shape, sourceX=sourceX,
                                                sourceY=sourceY, targetX=targetX, targetY=targetY,
                                                tension=tension, **kwds)


class facetTransform(VegaSchema):
    """facetTransform schema wrapper
    
    Mapping(required=[type])
    A special aggregate transform that organizes a data set into groups or "facets".
    
    Attributes
    ----------
    type : enum('facet')
    
    groupby : List(oneOf(string, signal))
        A list of fields to split the data into groups.
    summarize : oneOf(Mapping(required=[]), List(Mapping(required=[field, ops])))
    
    transform : transform
    
    """
    _schema = {'$ref': '#/defs/facetTransform'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, groupby=Undefined, summarize=Undefined, transform=Undefined,
                 **kwds):
        super(facetTransform, self).__init__(type=type, groupby=groupby, summarize=summarize,
                                             transform=transform, **kwds)


class filterTransform(VegaSchema):
    """filterTransform schema wrapper
    
    Mapping(required=[type, test])
    Filters elements from a data set to remove unwanted items.
    
    Attributes
    ----------
    test : string
        A string containing an expression (in JavaScript syntax) for the filter predicate.
    type : enum('filter')
    
    """
    _schema = {'$ref': '#/defs/filterTransform'}
    _rootschema = Root._schema

    def __init__(self, test=Undefined, type=Undefined, **kwds):
        super(filterTransform, self).__init__(test=test, type=type, **kwds)


class foldTransform(VegaSchema):
    """foldTransform schema wrapper
    
    Mapping(required=[type, fields])
    Collapse ("fold") one or more data properties into two properties.
    
    Attributes
    ----------
    fields : oneOf(List(oneOf(string, signal)), signal)
    
    type : enum('fold')
    
    output : Mapping(required=[])
        Rename the output data fields
    """
    _schema = {'$ref': '#/defs/foldTransform'}
    _rootschema = Root._schema

    def __init__(self, fields=Undefined, type=Undefined, output=Undefined, **kwds):
        super(foldTransform, self).__init__(fields=fields, type=type, output=output, **kwds)


class forceTransform(VegaSchema):
    """forceTransform schema wrapper
    
    Mapping(required=[type, links])
    Performs force-directed layout for network data.
    
    Attributes
    ----------
    links : string
        The name of the link (edge) data set.
    type : enum('force')
    
    active : signal
        A signal representing the active node.
    alpha : oneOf(float, signal)
        A "temperature" parameter that determines how much node positions are adjusted at 
        each step.
    charge : oneOf(float, string, signal)
        The strength of the charge each node exerts.
    chargeDistance : oneOf(float, signal)
        The maximum distance over which charge forces are applied.
    fixed : string
        The name of a datasource containing the IDs of nodes with fixed positions.
    friction : oneOf(float, signal)
        The strength of the friction force used to stabilize the layout.
    gravity : oneOf(float, signal)
        The strength of the pseudo-gravity force that pulls nodes towards the center of the 
        layout area.
    interactive : oneOf(boolean, signal)
        Enables an interactive force-directed layout.
    iterations : oneOf(float, signal)
        The number of iterations to run the force directed layout.
    linkDistance : oneOf(float, string, signal)
        Determines the length of edges, in pixels.
    linkStrength : oneOf(float, string, signal)
        Determines the tension of edges (the spring constant).
    output : Mapping(required=[])
        Rename the output data fields
    size : oneOf(List(oneOf(float, signal)), signal)
        The dimensions [width, height] of this force layout.
    theta : oneOf(float, signal)
        The theta parameter for the Barnes-Hut algorithm, which is used to compute charge 
        forces between nodes.
    """
    _schema = {'$ref': '#/defs/forceTransform'}
    _rootschema = Root._schema

    def __init__(self, links=Undefined, type=Undefined, active=Undefined, alpha=Undefined,
                 charge=Undefined, chargeDistance=Undefined, fixed=Undefined, friction=Undefined,
                 gravity=Undefined, interactive=Undefined, iterations=Undefined, linkDistance=Undefined,
                 linkStrength=Undefined, output=Undefined, size=Undefined, theta=Undefined, **kwds):
        super(forceTransform, self).__init__(links=links, type=type, active=active, alpha=alpha,
                                             charge=charge, chargeDistance=chargeDistance, fixed=fixed,
                                             friction=friction, gravity=gravity,
                                             interactive=interactive, iterations=iterations,
                                             linkDistance=linkDistance, linkStrength=linkStrength,
                                             output=output, size=size, theta=theta, **kwds)


class formulaTransform(VegaSchema):
    """formulaTransform schema wrapper
    
    Mapping(required=[type, field, expr])
    Extends data elements with new values according to a calculation formula.
    
    Attributes
    ----------
    expr : string
        A string containing an expression (in JavaScript syntax) for the formula.
    field : string
        The property name in which to store the computed formula value.
    type : enum('formula')
    
    """
    _schema = {'$ref': '#/defs/formulaTransform'}
    _rootschema = Root._schema

    def __init__(self, expr=Undefined, field=Undefined, type=Undefined, **kwds):
        super(formulaTransform, self).__init__(expr=expr, field=field, type=type, **kwds)


class geoTransform(VegaSchema):
    """geoTransform schema wrapper
    
    Mapping(required=[type, lon, lat])
    Performs a cartographic projection. Given longitude and latitude values, sets corresponding 
    x and y properties for a mark.
    
    Attributes
    ----------
    lat : oneOf(string, signal)
        The input latitude values.
    lon : oneOf(string, signal)
        The input longitude values.
    type : enum('geo')
    
    center : oneOf(List(oneOf(float, signal)), signal)
        The center of the projection.
    clipAngle : oneOf(float, signal)
        The clip angle of the projection.
    clipExtent : oneOf(float, signal)
        The clip extent of the projection.
    output : Mapping(required=[])
        Rename the output data fields
    precision : oneOf(float, signal)
        The desired precision of the projection.
    projection : oneOf(string, signal)
        The type of cartographic projection to use.
    rotate : oneOf(float, signal)
        The rotation of the projection.
    scale : oneOf(float, signal)
        The scale of the projection.
    translate : oneOf(List(oneOf(float, signal)), signal)
        The translation of the projection.
    """
    _schema = {'$ref': '#/defs/geoTransform'}
    _rootschema = Root._schema

    def __init__(self, lat=Undefined, lon=Undefined, type=Undefined, center=Undefined,
                 clipAngle=Undefined, clipExtent=Undefined, output=Undefined, precision=Undefined,
                 projection=Undefined, rotate=Undefined, scale=Undefined, translate=Undefined, **kwds):
        super(geoTransform, self).__init__(lat=lat, lon=lon, type=type, center=center,
                                           clipAngle=clipAngle, clipExtent=clipExtent, output=output,
                                           precision=precision, projection=projection, rotate=rotate,
                                           scale=scale, translate=translate, **kwds)


class geopathTransform(VegaSchema):
    """geopathTransform schema wrapper
    
    Mapping(required=[type])
    Creates paths for geographic regions, such as countries, states and counties.
    
    Attributes
    ----------
    type : enum('geopath')
    
    center : oneOf(List(oneOf(float, signal)), signal)
        The center of the projection.
    clipAngle : oneOf(float, signal)
        The clip angle of the projection.
    clipExtent : oneOf(float, signal)
        The clip extent of the projection.
    field : oneOf(string, signal)
        The data field containing GeoJSON Feature data.
    output : Mapping(required=[])
        Rename the output data fields
    precision : oneOf(float, signal)
        The desired precision of the projection.
    projection : oneOf(string, signal)
        The type of cartographic projection to use.
    rotate : oneOf(float, signal)
        The rotation of the projection.
    scale : oneOf(float, signal)
        The scale of the projection.
    translate : oneOf(List(oneOf(float, signal)), signal)
        The translation of the projection.
    """
    _schema = {'$ref': '#/defs/geopathTransform'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, center=Undefined, clipAngle=Undefined, clipExtent=Undefined,
                 field=Undefined, output=Undefined, precision=Undefined, projection=Undefined,
                 rotate=Undefined, scale=Undefined, translate=Undefined, **kwds):
        super(geopathTransform, self).__init__(type=type, center=center, clipAngle=clipAngle,
                                               clipExtent=clipExtent, field=field, output=output,
                                               precision=precision, projection=projection,
                                               rotate=rotate, scale=scale, translate=translate, **kwds)


class hierarchyTransform(VegaSchema):
    """hierarchyTransform schema wrapper
    
    Mapping(required=[type])
    
    Attributes
    ----------
    type : enum('hierarchy')
    
    children : oneOf(string, signal)
        The data field for the children node array
    field : oneOf(string, signal)
        The value for the area of each leaf-level node for partition layouts.
    mode : oneOf(enum('tidy', 'cluster', 'partition'), signal)
        The layout algorithm mode to use.
    nodesize : oneOf(List(oneOf(float, signal)), signal)
        Sets a fixed x,y size for each node (overrides the size parameter)
    orient : oneOf(enum('cartesian', 'radial'), signal)
        The layout orientation to use.
    output : Mapping(required=[])
        Rename the output data fields
    parent : oneOf(string, signal)
        The data field for the parent node
    size : oneOf(List(oneOf(float, signal)), signal)
        The dimensions of the tree layout
    sort : oneOf(List(oneOf(string, signal)), signal)
        A list of fields to use as sort criteria for sibling nodes.
    """
    _schema = {'$ref': '#/defs/hierarchyTransform'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, children=Undefined, field=Undefined, mode=Undefined,
                 nodesize=Undefined, orient=Undefined, output=Undefined, parent=Undefined,
                 size=Undefined, sort=Undefined, **kwds):
        super(hierarchyTransform, self).__init__(type=type, children=children, field=field, mode=mode,
                                                 nodesize=nodesize, orient=orient, output=output,
                                                 parent=parent, size=size, sort=sort, **kwds)


class imputeTransform(VegaSchema):
    """imputeTransform schema wrapper
    
    Mapping(required=[type, groupby, orderby, field])
    Performs imputation of missing values.
    
    Attributes
    ----------
    field : oneOf(string, signal)
        The data field to impute.
    groupby : oneOf(List(oneOf(string, signal)), signal)
        A list of fields to group the data into series.
    orderby : oneOf(List(oneOf(string, signal)), signal)
        A list of fields to determine ordering within series.
    type : enum('impute')
    
    method : oneOf(enum('value', 'mean', 'median', 'min', 'max'), signal)
        The imputation method to use.
    value : oneOf(float, string, boolean, None, signal)
        The value to use for missing data if the method is 'value'.
    """
    _schema = {'$ref': '#/defs/imputeTransform'}
    _rootschema = Root._schema

    def __init__(self, field=Undefined, groupby=Undefined, orderby=Undefined, type=Undefined,
                 method=Undefined, value=Undefined, **kwds):
        super(imputeTransform, self).__init__(field=field, groupby=groupby, orderby=orderby, type=type,
                                              method=method, value=value, **kwds)


class lookupTransform(VegaSchema):
    """lookupTransform schema wrapper
    
    Mapping(required=[type, on, as, keys])
    Extends a data set by looking up values in another data set.
    
    Attributes
    ----------
    keys : List(oneOf(string, signal))
        One or more fields in the primary data set to match against the secondary data set.
    on : string
        The name of the secondary data set on which to lookup values.
    type : enum('lookup')
    
    default : Mapping(required=[])
        The default value to use if a lookup match fails.
    onKey : oneOf(string, signal)
        The key field to lookup, or null for index-based lookup.
    """
    _schema = {'$ref': '#/defs/lookupTransform'}
    _rootschema = Root._schema

    def __init__(self, keys=Undefined, on=Undefined, type=Undefined, default=Undefined, onKey=Undefined,
                 **kwds):
        super(lookupTransform, self).__init__(keys=keys, on=on, type=type, default=default, onKey=onKey,
                                              **kwds)


class pieTransform(VegaSchema):
    """pieTransform schema wrapper
    
    Mapping(required=[type])
    Computes a pie chart layout.
    
    Attributes
    ----------
    type : enum('pie')
    
    endAngle : oneOf(float, signal)
    
    field : oneOf(string, signal)
        The data values to encode as angular spans. If this property is omitted, all pie 
        slices will have equal spans.
    output : Mapping(required=[])
        Rename the output data fields
    sort : oneOf(boolean, signal)
         If true, will sort the data prior to computing angles.
    startAngle : oneOf(float, signal)
    
    """
    _schema = {'$ref': '#/defs/pieTransform'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, endAngle=Undefined, field=Undefined, output=Undefined,
                 sort=Undefined, startAngle=Undefined, **kwds):
        super(pieTransform, self).__init__(type=type, endAngle=endAngle, field=field, output=output,
                                           sort=sort, startAngle=startAngle, **kwds)


class rankTransform(VegaSchema):
    """rankTransform schema wrapper
    
    Mapping(required=[type])
    Computes ascending rank scores for data tuples.
    
    Attributes
    ----------
    type : enum('rank')
    
    field : oneOf(string, signal)
        A key field to used to rank tuples. If undefined, tuples will be ranked in their 
        observed order.
    normalize : oneOf(boolean, signal)
        If true, values of the output field will lie in the range [0, 1].
    output : Mapping(required=[])
        Rename the output data fields
    """
    _schema = {'$ref': '#/defs/rankTransform'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, field=Undefined, normalize=Undefined, output=Undefined, **kwds):
        super(rankTransform, self).__init__(type=type, field=field, normalize=normalize, output=output,
                                            **kwds)


class sortTransform(VegaSchema):
    """sortTransform schema wrapper
    
    Mapping(required=[type, by])
    Sorts the values of a data set.
    
    Attributes
    ----------
    by : oneOf(string, List(string))
        A list of fields to use as sort criteria.
    type : enum('sort')
    
    """
    _schema = {'$ref': '#/defs/sortTransform'}
    _rootschema = Root._schema

    def __init__(self, by=Undefined, type=Undefined, **kwds):
        super(sortTransform, self).__init__(by=by, type=type, **kwds)


class stackTransform(VegaSchema):
    """stackTransform schema wrapper
    
    Mapping(required=[type, groupby, field])
    Computes layout values for stacked graphs, as in stacked bar charts or stream graphs.
    
    Attributes
    ----------
    field : oneOf(string, signal)
        The data field that determines the thickness/height of stacks.
    groupby : oneOf(List(oneOf(string, signal)), signal)
        A list of fields to split the data into groups (stacks).
    type : enum('stack')
    
    offset : oneOf(enum('zero', 'center', 'normalize'), signal)
        The baseline offset
    output : Mapping(required=[])
        Rename the output data fields
    sortby : oneOf(List(oneOf(string, signal)), signal)
        A list of fields to determine the sort order of stacks.
    """
    _schema = {'$ref': '#/defs/stackTransform'}
    _rootschema = Root._schema

    def __init__(self, field=Undefined, groupby=Undefined, type=Undefined, offset=Undefined,
                 output=Undefined, sortby=Undefined, **kwds):
        super(stackTransform, self).__init__(field=field, groupby=groupby, type=type, offset=offset,
                                             output=output, sortby=sortby, **kwds)


class treeifyTransform(VegaSchema):
    """treeifyTransform schema wrapper
    
    Mapping(required=[type, groupby])
    
    Attributes
    ----------
    groupby : oneOf(List(oneOf(string, signal)), signal)
        An ordered list of fields by which to group tuples into a tree.
    type : enum('treeify')
    
    output : Mapping(required=[])
        Rename the output data fields
    """
    _schema = {'$ref': '#/defs/treeifyTransform'}
    _rootschema = Root._schema

    def __init__(self, groupby=Undefined, type=Undefined, output=Undefined, **kwds):
        super(treeifyTransform, self).__init__(groupby=groupby, type=type, output=output, **kwds)


class treemapTransform(VegaSchema):
    """treemapTransform schema wrapper
    
    Mapping(required=[type])
    
    Attributes
    ----------
    type : enum('treemap')
    
    children : oneOf(string, signal)
        The data field for the children node array
    field : oneOf(string, signal)
        The values to use to determine the area of each leaf-level treemap cell.
    mode : oneOf(enum('squarify', 'slice', 'dice', 'slice-dice'), signal)
        The treemap layout algorithm to use.
    output : Mapping(required=[])
        Rename the output data fields
    padding : oneOf(float, List(oneOf(float, signal)), signal)
        he padding (in pixels) to provide around internal nodes in the treemap.
    parent : oneOf(string, signal)
        The data field for the parent node
    ratio : oneOf(float, signal)
        The target aspect ratio for the layout to optimize.
    round : oneOf(boolean, signal)
        If true, treemap cell dimensions will be rounded to integer pixels.
    size : oneOf(List(oneOf(float, signal)), signal)
        The dimensions of the treemap layout
    sort : oneOf(List(oneOf(string, signal)), signal)
        A list of fields to use as sort criteria for sibling nodes.
    sticky : oneOf(boolean, signal)
        If true, repeated runs of the treemap will use cached partition boundaries.
    """
    _schema = {'$ref': '#/defs/treemapTransform'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, children=Undefined, field=Undefined, mode=Undefined,
                 output=Undefined, padding=Undefined, parent=Undefined, ratio=Undefined,
                 round=Undefined, size=Undefined, sort=Undefined, sticky=Undefined, **kwds):
        super(treemapTransform, self).__init__(type=type, children=children, field=field, mode=mode,
                                               output=output, padding=padding, parent=parent,
                                               ratio=ratio, round=round, size=size, sort=sort,
                                               sticky=sticky, **kwds)


class voronoiTransform(VegaSchema):
    """voronoiTransform schema wrapper
    
    Mapping(required=[type])
    
    Attributes
    ----------
    type : enum('voronoi')
    
    clipExtent : oneOf(List(oneOf(List(oneOf(float, signal)), signal)), signal)
        The min and max points at which to clip the voronoi diagram.
    output : Mapping(required=[])
        Rename the output data fields
    x : oneOf(string, signal)
        The input x coordinates.
    y : oneOf(string, signal)
        The input y coordinates.
    """
    _schema = {'$ref': '#/defs/voronoiTransform'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, clipExtent=Undefined, output=Undefined, x=Undefined, y=Undefined,
                 **kwds):
        super(voronoiTransform, self).__init__(type=type, clipExtent=clipExtent, output=output, x=x,
                                               y=y, **kwds)


class wordcloudTransform(VegaSchema):
    """wordcloudTransform schema wrapper
    
    Mapping(required=[type])
    
    Attributes
    ----------
    type : enum('wordcloud')
    
    font : oneOf(string, oneOf(Mapping(required=[field]), Mapping(required=[value])), signal)
        The font face to use for a word.
    fontScale : oneOf(None, List(oneOf(float, signal)))
        The minimum and maximum scaled font sizes, or null to prevent scaling.
    fontSize : oneOf(float, oneOf(Mapping(required=[field]), Mapping(required=[value])), string,
     signal)
        The font size to use for a word.
    fontStyle : oneOf(string, oneOf(Mapping(required=[field]), Mapping(required=[value])), 
    signal)
        The font style to use for a word.
    fontWeight : oneOf(string, oneOf(Mapping(required=[field]), Mapping(required=[value])), 
    signal)
        The font weight to use for a word.
    output : Mapping(required=[])
        Rename the output data fields
    padding : oneOf(float, oneOf(Mapping(required=[field]), Mapping(required=[value])), signal)
        The padding around each word.
    rotate : oneOf(float, string, oneOf(Mapping(required=[field]), Mapping(required=[value])), 
    signal)
        The field or number to set the roration angle (in degrees).
    size : oneOf(List(oneOf(float, signal)), signal)
        The dimensions of the wordcloud layout
    spiral : oneOf(enum('archimedean', 'rectangular'), oneOf(Mapping(required=[field]), 
    Mapping(required=[value])), signal)
        The type of spiral used for positioning words, either 'archimedean' or 
        'rectangular'.
    text : oneOf(string, oneOf(Mapping(required=[field]), Mapping(required=[value])), signal)
        The field containing the text to use for each word.
    """
    _schema = {'$ref': '#/defs/wordcloudTransform'}
    _rootschema = Root._schema

    def __init__(self, type=Undefined, font=Undefined, fontScale=Undefined, fontSize=Undefined,
                 fontStyle=Undefined, fontWeight=Undefined, output=Undefined, padding=Undefined,
                 rotate=Undefined, size=Undefined, spiral=Undefined, text=Undefined, **kwds):
        super(wordcloudTransform, self).__init__(type=type, font=font, fontScale=fontScale,
                                                 fontSize=fontSize, fontStyle=fontStyle,
                                                 fontWeight=fontWeight, output=output, padding=padding,
                                                 rotate=rotate, size=size, spiral=spiral, text=text,
                                                 **kwds)


class transform(VegaSchema):
    """transform schema wrapper
    
    List(oneOf(aggregateTransform, binTransform, crossTransform, countpatternTransform, 
    linkpathTransform, facetTransform, filterTransform, foldTransform, forceTransform, 
    formulaTransform, geoTransform, geopathTransform, hierarchyTransform, imputeTransform, 
    lookupTransform, pieTransform, rankTransform, sortTransform, stackTransform, 
    treeifyTransform, treemapTransform, voronoiTransform, wordcloudTransform))
    """
    _schema = {'$ref': '#/defs/transform'}
    _rootschema = Root._schema

    def __init__(self, *args):
        super(transform, self).__init__(*args)


class scale(VegaSchema):
    """scale schema wrapper
    
    allOf(Mapping(required=[name]), oneOf(Mapping(required=[type]), Mapping(required=[type]), 
    anyOf(Mapping(required=[]), Mapping(required=[type]))))
    """
    _schema = {'$ref': '#/defs/scale'}
    _rootschema = Root._schema

    def __init__(self, name=Undefined, domain=Undefined, domainMax=Undefined, domainMin=Undefined,
                 rangeMax=Undefined, rangeMin=Undefined, reverse=Undefined, round=Undefined,
                 type=Undefined, **kwds):
        super(scale, self).__init__(name=name, domain=domain, domainMax=domainMax, domainMin=domainMin,
                                    rangeMax=rangeMax, rangeMin=rangeMin, reverse=reverse, round=round,
                                    type=type, **kwds)


class operand(VegaSchema):
    """operand schema wrapper
    
    oneOf(Mapping(required=[value]), Mapping(required=[arg]), signal, 
    Mapping(required=[predicate]))
    """
    _schema = {'$ref': '#/refs/operand'}
    _rootschema = Root._schema

    def __init__(self, *args, **kwds):
        super(operand, self).__init__(*args, **kwds)


class field(VegaSchema):
    """field schema wrapper
    
    oneOf(string, oneOf(signal, Mapping(required=[datum]), Mapping(required=[group]), 
    Mapping(required=[parent])))
    """
    _schema = {'$ref': '#/refs/field'}
    _rootschema = Root._schema

    def __init__(self, *args, **kwds):
        super(field, self).__init__(*args, **kwds)


class scale(VegaSchema):
    """scale schema wrapper
    
    oneOf(field, Mapping(required=[name]))
    """
    _schema = {'$ref': '#/refs/scale'}
    _rootschema = Root._schema

    def __init__(self, *args, **kwds):
        super(scale, self).__init__(*args, **kwds)


class stringModifiers(VegaSchema):
    """stringModifiers schema wrapper
    
    Mapping(required=[])
    
    Attributes
    ----------
    scale : scale
    
    """
    _schema = {'$ref': '#/refs/stringModifiers'}
    _rootschema = Root._schema

    def __init__(self, scale=Undefined, **kwds):
        super(stringModifiers, self).__init__(scale=scale, **kwds)


class numberModifiers(VegaSchema):
    """numberModifiers schema wrapper
    
    Mapping(required=[])
    
    Attributes
    ----------
    mult : float
    
    offset : float
    
    scale : scale
    
    """
    _schema = {'$ref': '#/refs/numberModifiers'}
    _rootschema = Root._schema

    def __init__(self, mult=Undefined, offset=Undefined, scale=Undefined, **kwds):
        super(numberModifiers, self).__init__(mult=mult, offset=offset, scale=scale, **kwds)


class value(VegaSchema):
    """value schema wrapper
    
    oneOf(Mapping(required=[rule]), List(allOf(rule, allOf(stringModifiers, oneOf(signal, 
    Mapping(required=[value]), Mapping(required=[field]), Mapping(required=[band]))))), 
    allOf(stringModifiers, oneOf(signal, Mapping(required=[value]), Mapping(required=[field]), 
    Mapping(required=[band]))))
    """
    _schema = {'$ref': '#/refs/value'}
    _rootschema = Root._schema

    def __init__(self, *args, **kwds):
        super(value, self).__init__(*args, **kwds)


class numberValue(VegaSchema):
    """numberValue schema wrapper
    
    oneOf(Mapping(required=[rule]), List(allOf(rule, allOf(numberModifiers, oneOf(signal, 
    Mapping(required=[value]), Mapping(required=[field]), Mapping(required=[band]))))), 
    allOf(numberModifiers, oneOf(signal, Mapping(required=[value]), Mapping(required=[field]), 
    Mapping(required=[band]))))
    """
    _schema = {'$ref': '#/refs/numberValue'}
    _rootschema = Root._schema

    def __init__(self, *args, **kwds):
        super(numberValue, self).__init__(*args, **kwds)


class stringValue(VegaSchema):
    """stringValue schema wrapper
    
    oneOf(Mapping(required=[rule]), List(allOf(rule, allOf(stringModifiers, oneOf(signal, 
    Mapping(required=[value]), Mapping(required=[field]), Mapping(required=[band]), 
    Mapping(required=[template]))))), allOf(stringModifiers, oneOf(signal, 
    Mapping(required=[value]), Mapping(required=[field]), Mapping(required=[band]), 
    Mapping(required=[template]))))
    """
    _schema = {'$ref': '#/refs/stringValue'}
    _rootschema = Root._schema

    def __init__(self, *args, **kwds):
        super(stringValue, self).__init__(*args, **kwds)


class booleanValue(VegaSchema):
    """booleanValue schema wrapper
    
    oneOf(Mapping(required=[rule]), List(allOf(rule, allOf(stringModifiers, oneOf(signal, 
    Mapping(required=[value]), Mapping(required=[field]), Mapping(required=[band]))))), 
    allOf(stringModifiers, oneOf(signal, Mapping(required=[value]), Mapping(required=[field]), 
    Mapping(required=[band]))))
    """
    _schema = {'$ref': '#/refs/booleanValue'}
    _rootschema = Root._schema

    def __init__(self, *args, **kwds):
        super(booleanValue, self).__init__(*args, **kwds)


class arrayValue(VegaSchema):
    """arrayValue schema wrapper
    
    oneOf(Mapping(required=[rule]), List(allOf(rule, allOf(stringModifiers, oneOf(signal, 
    Mapping(required=[value]), Mapping(required=[field]), Mapping(required=[band]))))), 
    allOf(stringModifiers, oneOf(signal, Mapping(required=[value]), Mapping(required=[field]), 
    Mapping(required=[band]))))
    """
    _schema = {'$ref': '#/refs/arrayValue'}
    _rootschema = Root._schema

    def __init__(self, *args, **kwds):
        super(arrayValue, self).__init__(*args, **kwds)


class colorValue(VegaSchema):
    """colorValue schema wrapper
    
    oneOf(stringValue, Mapping(required=[r, g, b]), Mapping(required=[h, s, l]), 
    Mapping(required=[l, a, b]), Mapping(required=[h, c, l]))
    """
    _schema = {'$ref': '#/refs/colorValue'}
    _rootschema = Root._schema

    def __init__(self, *args, **kwds):
        super(colorValue, self).__init__(*args, **kwds)


class signal(VegaSchema):
    """signal schema wrapper
    
    Mapping(required=[signal])
    
    Attributes
    ----------
    signal : string
    
    """
    _schema = {'$ref': '#/refs/signal'}
    _rootschema = Root._schema

    def __init__(self, signal=Undefined, **kwds):
        super(signal, self).__init__(signal=signal, **kwds)


class scopedScale(VegaSchema):
    """scopedScale schema wrapper
    
    oneOf(string, Mapping(required=[name]))
    """
    _schema = {'$ref': '#/refs/scopedScale'}
    _rootschema = Root._schema

    def __init__(self, *args, **kwds):
        super(scopedScale, self).__init__(*args, **kwds)


class data(VegaSchema):
    """data schema wrapper
    
    Mapping(required=[])
    
    Attributes
    ----------
    data : oneOf(string, Mapping(required=[fields]))
    
    field : oneOf(string, List(string), Mapping(required=[parent]), 
    List(Mapping(required=[parent])))
    
    sort : oneOf(boolean, Mapping(required=[]))
    
    """
    _schema = {'$ref': '#/refs/data'}
    _rootschema = Root._schema

    def __init__(self, data=Undefined, field=Undefined, sort=Undefined, **kwds):
        super(data, self).__init__(data=data, field=field, sort=sort, **kwds)




"""
Stacked Bar Chart
-----------------

This example shows how to make a stacked bar chart of the weather type in Seattle from 2012 to 2015 by month.
"""

import altair as alt
from vega_datasets import data

weather = data.seattle_weather()

chart = alt.Chart(weather).mark_bar().encode(
    alt.Color('weather:N',
        legend=alt.Legend(title='Weather type'),
        scale=alt.Scale(
            domain=['sun', 'fog', 'drizzle', 'rain', 'snow'],
            range=['#e7ba42', '#c7c7c7', '#aec7e8', '#1f77b4', '#9467bd']
        ),
    ),
    alt.X('date:N',
        axis=alt.Axis(title='Month of the Year'),
        timeUnit='month',
    ),
    y='count(*):Q',
)



"""
Altair Plot Sphinx Extension
============================

This extension provides a means of inserting live-rendered Altair plots within
sphinx documentation. There are two directives defined: ``altair-setup`` and
``altiar-plot``. ``altair-setup`` code is used to set-up various options
prior to running the plot code. For example::

    .. altair-plot::
        :output: none

        from altair import *
        import pandas as pd
        data = pd.DataFrame({'a': list('CCCDDDEEE'),
                             'b': [2, 7, 4, 1, 2, 6, 8, 4, 7]})

    .. altair-plot::

        Chart(data).mark_point().encode(
            x='a',
            y='b'
        )

In the case of the ``altair-plot`` code, the *last statement* of the code-block
should contain the chart object you wish to be rendered.

Options
-------
The directives have the following options::

    .. altair-plot::
        :namespace:  # specify a plotting namespace that is persistent within the doc
        :hide-code:  # if set, then hide the code and only show the plot
        :code-below:  # if set, then code is below rather than above the figure
        :output:  [plot|repr|stdout|none]
        :alt: text  # Alternate text when plot cannot be rendered
        :links: editor source export  # specify one or more of these options
        :chart-var-name: chart  # name of variable in namespace containing output


Additionally, this extension introduces a global configuration
``altairplot_links``, set in your ``conf.py`` which is a dictionary
of links that will appear below plots, unless the ``:links:`` option
again overrides it. It should look something like this::

    # conf.py
    # ...
    altairplot_links = {'editor': True, 'source': True, 'export': True}
    # ...

If this configuration is not specified, all are set to True.
"""

import contextlib
import io
import os
import json
import warnings

import jinja2

from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.parsers.rst.directives import flag, unchanged

from sphinx.locale import _
from sphinx import addnodes, directives
from sphinx.util.nodes import set_source_info

import altair as alt
from .utils import exec_then_eval

# These default URLs can be changed in conf.py; see setup() below.
VEGA_JS_URL_DEFAULT = "https://cdn.jsdelivr.net/npm/vega@3"
VEGALITE_JS_URL_DEFAULT = "https://cdn.jsdelivr.net/npm/vega-lite@2"
VEGAEMBED_JS_URL_DEFAULT = "https://cdn.jsdelivr.net/npm/vega-embed@3"


VGL_TEMPLATE = jinja2.Template("""
<div id="{{ div_id }}">
<script>
  // embed when document is loaded, to ensure vega library is available
  // this works on all modern browsers, except IE8 and older
  document.addEventListener("DOMContentLoaded", function(event) {
      var spec = {{ spec }};
      var opt = {
        "mode": "{{ mode }}",
        "renderer": "{{ renderer }}",
        "actions": {{ actions}}
      };
      vegaEmbed('#{{ div_id }}', {{ spec }}).catch(console.err);
  });
</script>
</div>
""")


class altair_plot(nodes.General, nodes.Element):
    pass


def purge_altair_namespaces(app, env, docname):
    if not hasattr(env, '_altair_namespaces'):
        return
    env._altair_namespaces.pop(docname, {})


DEFAULT_ALTAIRPLOT_LINKS = {'editor': True, 'source': True, 'export': True}


def validate_links(links):
    if links.strip().lower() == 'none':
        return {}

    links = links.strip().split()
    diff = set(links) - set(DEFAULT_ALTAIRPLOT_LINKS.keys())
    if diff:
        raise ValueError("Following links are invalid: {0}".format(list(diff)))
    return dict((link, link in links) for link in DEFAULT_ALTAIRPLOT_LINKS)


def validate_output(output):
    output = output.strip().lower()
    if output not in ['plot', 'repr', 'stdout', 'none']:
        raise ValueError(":output: flag must be one of [plot|repr|stdout|none]")
    return output


class AltairPlotDirective(Directive):
    has_content = True

    option_spec = {'hide-code': flag,
                   'code-below': flag,
                   'namespace': unchanged,
                   'output': validate_output,
                   'alt': unchanged,
                   'links': validate_links,
                   'chart-var-name': unchanged}

    def run(self):
        env = self.state.document.settings.env
        app = env.app

        show_code = 'hide-code' not in self.options
        code_below = 'code-below' in self.options

        if not hasattr(env, '_altair_namespaces'):
            env._altair_namespaces = {}
        namespace_id = self.options.get('namespace', 'default')
        namespace = env._altair_namespaces\
                        .setdefault(env.docname, {})\
                        .setdefault(namespace_id, {})

        code = '\n'.join(self.content)

        if show_code:
            source_literal = nodes.literal_block(code, code)
            source_literal['language'] = 'python'

        #get the name of the source file we are currently processing
        rst_source = self.state_machine.document['source']
        rst_dir = os.path.dirname(rst_source)
        rst_filename = os.path.basename(rst_source)

        # use the source file name to construct a friendly target_id
        serialno = env.new_serialno('altair-plot')
        rst_base = rst_filename.replace('.', '-')
        div_id = "{0}-altair-plot-{1}".format(rst_base, serialno)
        target_id = "{0}-altair-source-{1}".format(rst_base, serialno)
        target_node = nodes.target('', '', ids=[target_id])

        # create the node in which the plot will appear;
        # this will be processed by html_visit_altair_plot
        plot_node = altair_plot()
        plot_node['target_id'] = target_id
        plot_node['div_id'] = div_id
        plot_node['code'] = code
        plot_node['namespace'] = namespace
        plot_node['relpath'] = os.path.relpath(rst_dir, env.srcdir)
        plot_node['rst_source'] = rst_source
        plot_node['rst_lineno'] = self.lineno
        plot_node['links'] = self.options.get('links', app.builder.config.altairplot_links)
        plot_node['output'] = self.options.get('output', 'plot')
        plot_node['chart-var-name'] = self.options.get('chart-var-name', None)

        if 'alt' in self.options:
            plot_node['alt'] = self.options['alt']

        result = [target_node]

        if code_below:
            result += [plot_node]
        if show_code:
            result += [source_literal]
        if not code_below:
            result += [plot_node]

        return result


def html_visit_altair_plot(self, node):
    # Execute the code, saving output and namespace
    namespace = node['namespace']
    try:
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            chart = exec_then_eval(node['code'], namespace)
        stdout = f.getvalue()
    except Exception as e:
        warnings.warn("altair-plot: {0}:{1} Code Execution failed:"
                      "{2}: {3}".format(node['rst_source'], node['rst_lineno'],
                                        e.__class__.__name__, str(e)))
        raise nodes.SkipNode

    chart_name = node['chart-var-name']
    if chart_name is not None:
        if chart_name not in namespace:
            raise ValueError("chart-var-name='{0}' not present in namespace"
                             "".format(chart_name))
        chart = namespace[chart_name]


    output = node['output']

    if output == 'none':
        raise nodes.SkipNode
    elif output == 'stdout':
        if not stdout:
            raise nodes.SkipNode
        else:
            output_literal = nodes.literal_block(stdout, stdout)
            output_literal['language'] = 'none'
            node.extend([output_literal])
            self.visit_admonition(node)
    elif output == 'repr':
        if chart is None:
            raise nodes.SkipNode
        else:
            rep = '    ' + repr(chart).replace('\n', '\n    ')
            repr_literal = nodes.literal_block(rep, rep)
            repr_literal['language'] = 'none'
            node.extend([repr_literal])
            self.visit_admonition(node)
    elif output == 'plot':
        if isinstance(chart, alt.TopLevelMixin):
            # Last line should be a chart; convert to spec dict
            spec = chart.to_dict()
            actions = node['links']

            # TODO: add an option to save spects to file & load from there.
            # TODO: add renderer option

            # Write spec to a *.vl.json file
            # dest_dir = os.path.join(self.builder.outdir, node['relpath'])
            # if not os.path.exists(dest_dir):
            #     os.makedirs(dest_dir)
            # filename = "{0}.vl.json".format(node['target_id'])
            # dest_path = os.path.join(dest_dir, filename)
            # with open(dest_path, 'w') as f:
            #     json.dump(spec, f)

            # Pass relevant info into the template and append to the output
            html = VGL_TEMPLATE.render(div_id=node['div_id'],
                                       spec=json.dumps(spec),
                                       mode='vega-lite',
                                       renderer='canvas',
                                       actions=json.dumps(actions))
            self.body.append(html)
        else:
            warnings.warn('altair-plot: {0}:{1} Malformed block. Last line of '
                          'code block should define a valid altair Chart object.'
                          ''.format(node['rst_source'], node['rst_lineno']))
        raise nodes.SkipNode


def generic_visit_altair_plot(self, node):
    # TODO: generate PNGs and insert them here
    if 'alt' in node.attributes:
        self.body.append(_('[ graph: %s ]') % node['alt'])
    else:
        self.body.append(_('[ graph ]'))
    raise nodes.SkipNode


def depart_altair_plot(self, node):
    return


def builder_inited(app):
    app.add_javascript(app.config.altairplot_vega_js_url)
    app.add_javascript(app.config.altairplot_vegalite_js_url)
    app.add_javascript(app.config.altairplot_vegaembed_js_url)


def setup(app):
    setup.app = app
    setup.config = app.config
    setup.confdir = app.confdir

    app.add_config_value('altairplot_links', DEFAULT_ALTAIRPLOT_LINKS, 'env')

    app.add_config_value('altairplot_vega_js_url', VEGA_JS_URL_DEFAULT, 'html')
    app.add_config_value('altairplot_vegalite_js_url', VEGALITE_JS_URL_DEFAULT, 'html')
    app.add_config_value('altairplot_vegaembed_js_url', VEGAEMBED_JS_URL_DEFAULT, 'html')

    app.add_directive('altair-plot', AltairPlotDirective)

    app.add_stylesheet('altair-plot.css')

    app.add_node(altair_plot,
                 html=(html_visit_altair_plot, depart_altair_plot),
                 latex=(generic_visit_altair_plot, depart_altair_plot),
                 texinfo=(generic_visit_altair_plot, depart_altair_plot),
                 text=(generic_visit_altair_plot, depart_altair_plot),
                 man=(generic_visit_altair_plot, depart_altair_plot))

    app.connect('env-purge-doc', purge_altair_namespaces)
    app.connect('builder-inited', builder_inited)

    return {'version': '0.1'}
